from asyncio import gather
from base64 import b64decode
from collections.abc import AsyncGenerator, Callable, Coroutine, Mapping, Sequence
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from types import TracebackType
from typing import Any, Self, cast, final

from haiway import as_dict, as_list
from mcp import ClientSession, GetPromptResult, ListToolsResult, StdioServerParameters, stdio_client
from mcp import Tool as MCPTool
from mcp.client.sse import sse_client
from mcp.types import (
    BlobResourceContents,
    CallToolResult,
    EmbeddedResource,
    ListResourcesResult,
    ReadResourceResult,
    TextResourceContents,
)
from mcp.types import ImageContent as MCPImageContent
from mcp.types import TextContent as MCPTextContent
from pydantic import AnyUrl

from draive.lmm import LMMCompletion, LMMContextElement, LMMInput
from draive.lmm.types import LMMToolError
from draive.multimodal import MediaContent, MultimodalContent, TextContent, validated_media_type
from draive.parameters import BasicValue, ParametersSpecification
from draive.parameters.model import DataModel
from draive.prompts import Prompt, PromptDeclaration, PromptDeclarationArgument, PromptRepository
from draive.resources import Resource, ResourceContent, ResourceDeclaration, ResourceRepository
from draive.tools import AnyTool, ExternalToolbox, Tool, Toolbox

__all__ = [
    "MCPClient",
]


@final
class MCPClient:
    @classmethod
    def stdio(
        cls,
        *,
        command: str,
        args: Sequence[str] | None = None,
        env: dict[str, str] | None = None,
    ) -> Self:
        @asynccontextmanager
        async def mcp_stdio_session() -> AsyncGenerator[ClientSession]:
            async with stdio_client(
                StdioServerParameters(
                    command=command,
                    args=as_list(args) if args else [],
                    env=env,
                )
            ) as (read, write):
                async with ClientSession(read, write) as session:
                    yield session

        return cls(mcp_stdio_session())

    @classmethod
    def sse(
        cls,
        *,
        url: str,
        headers: dict[str, Any] | None = None,
        timeout: float = 5,
        sse_read_timeout: float = 60 * 5,
    ) -> Self:
        @asynccontextmanager
        async def mcp_sse_session() -> AsyncGenerator[ClientSession]:
            async with sse_client(
                url=url,
                headers=headers,
                timeout=timeout,
                sse_read_timeout=sse_read_timeout,
            ) as (read, write):
                async with ClientSession(read, write) as session:
                    yield session

        return cls(mcp_sse_session())

    def __init__(
        self,
        session_manager: AbstractAsyncContextManager[ClientSession],
        /,
    ) -> None:
        self._session_manager: AbstractAsyncContextManager[ClientSession] = session_manager
        self._session: ClientSession

    async def resources_list(
        self,
        **extra: Any,
    ) -> Sequence[ResourceDeclaration]:
        assert hasattr(  # nosec: B101
            self,
            "_session",
        ), "MCPClient has to be initialized throug async context entering"

        result: ListResourcesResult = await self._session.list_resources()
        # TODO: in theory there are some extra elements like pagination
        # we are not supporting it as for now (how to pass cursor anyways?)
        return [
            ResourceDeclaration(
                uri=resource.uri.unicode_string(),
                name=resource.name,
                description=resource.description,
                mime_type=resource.mimeType,
            )
            for resource in result.resources
        ]

    async def resource_fetch(
        self,
        uri: str,
        **extra: Any,
    ) -> Resource | None:
        assert hasattr(  # nosec: B101
            self,
            "_session",
        ), "MCPClient has to be initialized throug async context entering"
        result: ReadResourceResult = await self._session.read_resource(uri=AnyUrl(uri))

        match [_convert_resource_content(element) for element in result.contents]:
            case [resource]:
                # if there is only a single element return it directly
                return resource

            case resources:
                return Resource(
                    uri=uri,
                    name="",  # TODO: resource name?
                    description=None,
                    content=resources,
                )

    async def prompts_list(
        self,
        **extra: Any,
    ) -> Sequence[PromptDeclaration]:
        assert hasattr(  # nosec: B101
            self,
            "_session",
        ), "MCPClient has to be initialized throug async context entering"

        # TODO: in theory there are some extra elements like pagination
        # we are not supporting it as for now (how to pass cursor anyways?)
        return tuple(
            PromptDeclaration(
                name=prompt.name,
                description=prompt.description,
                arguments=tuple(
                    PromptDeclarationArgument(
                        name=argument.name,
                        specification={
                            "type": "string",
                        }
                        if argument.description is None
                        else {
                            "type": "string",
                            "description": argument.description,
                        },
                        required=argument.required if argument.required is not None else True,
                    )
                    for argument in prompt.arguments
                )
                if prompt.arguments
                else (),
            )
            for prompt in (await self._session.list_prompts()).prompts
        )

    async def prompt_fetch(
        self,
        name: str,
        *,
        arguments: Mapping[str, str] | None,
        **extra: Any,
    ) -> Prompt:
        assert hasattr(  # nosec: B101
            self, "_session"
        ), "MCPClient has to be initialized throug async context entering"

        fetched_prompt: GetPromptResult = await self._session.get_prompt(
            name=name,
            arguments=as_dict(arguments) if arguments is not None else None,
        )

        prompt_context: list[LMMContextElement] = []
        for element in fetched_prompt.messages:
            match element.role:
                case "user":
                    prompt_context.append(LMMInput.of(await _convert_content(element.content)))

                case "assistant":
                    prompt_context.append(LMMCompletion.of(await _convert_content(element.content)))

        return Prompt(
            name=name,
            description=fetched_prompt.description,
            content=prompt_context,
        )

    async def toolbox_fetch(
        self,
        *,
        suggest: bool | None = None,
        repeated_calls_limit: int | None = None,
        **extra: Any,
    ) -> Toolbox:
        tools: ListToolsResult = await self._session.list_tools()

        return Toolbox.of(
            *[_convert_tool(tool, self.tool_call) for tool in tools.tools],
            suggest=suggest,
            repeated_calls_limit=repeated_calls_limit,
        )

    async def tool_call(
        self,
        name: str,
        arguments: Mapping[str, BasicValue],
    ) -> MultimodalContent:
        # TODO: FIXME: linter
        result: CallToolResult = await self._session.call_tool(  # pyright: ignore
            name=name,
            arguments=as_dict(arguments),
        )
        content: MultimodalContent = MultimodalContent.of(
            *await gather(*(_convert_content(part) for part in result.content))
        )

        if result.isError:
            raise LMMToolError(
                f"Remote tool {name} failed",
                content=content,
            )

        return content

    async def __aenter__(self) -> tuple[ResourceRepository, PromptRepository, ExternalToolbox]:
        self._session = await self._session_manager.__aenter__()
        await self._session.initialize()

        return (
            ResourceRepository(
                list=self.resources_list,
                fetch=self.resource_fetch,
            ),
            PromptRepository(
                list=self.prompts_list,
                fetch=self.prompt_fetch,
            ),
            ExternalToolbox(
                fetch=self.toolbox_fetch,
            ),
        )

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._session_manager.__aexit__(
            exc_type,
            exc_val,
            exc_tb,
        )
        del self._session


def _convert_resource_content(
    resource: TextResourceContents | BlobResourceContents,
    /,
) -> Resource:
    match resource:
        case TextResourceContents():
            return Resource(
                uri=resource.uri.unicode_string(),
                name="",  # TODO: resource name?
                description=None,
                content=ResourceContent(
                    blob=resource.text.encode(),
                    mime_type=resource.mimeType or "text/plain",
                ),
            )

        case BlobResourceContents():
            return Resource(
                uri=resource.uri.unicode_string(),
                name="",  # TODO: resource name?
                description=None,
                content=ResourceContent(
                    blob=b64decode(resource.blob),
                    mime_type=resource.mimeType or "text/plain",
                ),
            )


async def _convert_content(
    content: MCPTextContent | MCPImageContent | EmbeddedResource,
    /,
) -> MultimodalContent:
    match content:
        case MCPTextContent() as text:
            return MultimodalContent.of(text.text)

        case MCPImageContent() as image:
            return MultimodalContent.of(
                MediaContent.base64(
                    image.data,
                    media=validated_media_type(image.mimeType),
                )
            )

        case EmbeddedResource() as resource:
            match resource.resource:
                case TextResourceContents() as text:
                    return MultimodalContent.of(TextContent(text=text.text))

                case BlobResourceContents() as blob:
                    match blob.mimeType:
                        case None:
                            raise NotImplementedError(
                                "Unsuppoprted MCP prompt message element - missing mime!"
                            )

                        case "text/plain":
                            return MultimodalContent.of(
                                TextContent(text=b64decode(blob.blob).decode())
                            )

                        case "application/json":
                            return MultimodalContent.of(
                                DataModel.from_json(b64decode(blob.blob).decode())
                            )

                        case other:
                            # try to match supported media or raise an exception
                            return MultimodalContent.of(
                                MediaContent.data(
                                    b64decode(blob.blob),
                                    media=validated_media_type(other),
                                )
                            )


def _convert_tool(
    mcp_tool: MCPTool,
    tool_call: Callable[[str, Mapping[str, BasicValue]], Coroutine[None, None, MultimodalContent]],
) -> AnyTool:
    name: str = mcp_tool.name

    async def remote_call(**arguments: Any) -> MultimodalContent:
        return await tool_call(
            name,
            arguments,
        )

    return Tool(
        name=name,
        description=mcp_tool.description,
        specification=cast(ParametersSpecification, mcp_tool.inputSchema),
        function=remote_call,
        availability_check=None,
        format_result=_format_tool_result,
        format_failure=_format_tool_failure,
        direct_result=False,
    )


def _format_tool_result(
    result: Any,
) -> MultimodalContent:
    assert isinstance(result, MultimodalContent)  # nosec: B101
    return result


def _format_tool_failure(
    exception: Exception,
) -> MultimodalContent:
    if isinstance(exception, LMMToolError):
        return exception.content

    return MultimodalContent.of("ERROR")



from pathlib import Path
from typing import Any

from haiway import asynchronous
from tokenizers import Tokenizer

from draive.tokenization import Tokenization

__all__ = [
    "anthropic_tokenizer",
]


async def anthropic_tokenizer() -> Tokenization:
    tokenizer: Tokenizer = await _tokenizer()

    def anthropic_tokenize_text(
        text: str,
        **extra: Any,
    ) -> list[int]:
        return tokenizer.encode(text).ids  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]

    return Tokenization(tokenize_text=anthropic_tokenize_text)


@asynchronous
def _tokenizer() -> Tokenizer:
    return Tokenizer.from_file(str(Path(__file__).parent / "tokenizer.json"))  # pyright: ignore



from collections.abc import Mapping
from typing import Any, Literal, overload

from haiway import ctx

from draive.instructions.state import InstructionsRepository
from draive.instructions.types import Instruction, InstructionDeclaration, MissingInstruction

__all__ = [
    "fetch_instruction",
]


@overload
async def fetch_instruction(
    reference: InstructionDeclaration | str,
    /,
    *,
    default: Instruction | str | None = None,
    arguments: Mapping[str, str] | None = None,
    **extra: Any,
) -> Instruction | None: ...


@overload
async def fetch_instruction(
    reference: InstructionDeclaration | str,
    /,
    *,
    default: Instruction | str,
    arguments: Mapping[str, str] | None = None,
    **extra: Any,
) -> Instruction: ...


@overload
async def fetch_instruction(
    reference: InstructionDeclaration | str,
    /,
    *,
    default: Instruction | str | None = None,
    arguments: Mapping[str, str] | None = None,
    required: Literal[True],
    **extra: Any,
) -> Instruction: ...


async def fetch_instruction(
    reference: InstructionDeclaration | str,
    /,
    *,
    default: Instruction | str | None = None,
    arguments: Mapping[str, str] | None = None,
    required: bool = True,
    **extra: Any,
) -> Instruction | None:
    name: str = reference if isinstance(reference, str) else reference.name

    match await ctx.state(InstructionsRepository).fetch(
        name,
        arguments=arguments,
        **extra,
    ):
        case None:
            match default:
                case None:
                    if required:
                        raise MissingInstruction(f"Missing instruction: '{name}'")

                    else:
                        return None

                case Instruction() as instruction:
                    return instruction

                case str() as text:
                    return Instruction.of(
                        text,
                        name=name,
                        description=reference.description
                        if isinstance(reference, InstructionDeclaration)
                        else None,
                        **(arguments if arguments is not None else {}),
                    )

        case instruction:
            return instruction



from typing import Any

from haiway import cache
from tiktoken import Encoding, encoding_for_model

from draive.tokenization import Tokenization

__all__ = [
    "openai_tokenizer",
]


def openai_tokenizer(
    model_name: str,
    /,
) -> Tokenization:
    encoding: Encoding = _encoding(model_name=model_name)

    def openai_tokenize_text(
        text: str,
        **extra: Any,
    ) -> list[int]:
        return encoding.encode(text=text)

    return Tokenization(tokenize_text=openai_tokenize_text)


@cache(limit=4)
def _encoding(model_name: str) -> Encoding:
    return encoding_for_model(model_name=model_name)



import json
from asyncio import gather
from collections.abc import AsyncIterator, Sequence
from http import HTTPStatus
from itertools import chain
from types import TracebackType
from typing import Any, ClassVar, Literal, Self, cast, final, overload

from haiway import getenv_str, not_missing
from httpx import AsyncClient, Response

from draive.mistral.config import MistralChatConfig, MistralEmbeddingConfig
from draive.mistral.models import (
    ChatCompletionResponse,
    ChatCompletionStreamResponse,
    ChatMessage,
    EmbeddingResponse,
)
from draive.mistral.types import MistralException
from draive.parameters import DataModel

__all__ = [
    "MistralClient",
]


@final
class MistralClient:
    _SHARED: ClassVar[Self]

    @classmethod
    def shared(cls) -> Self:
        if shared := getattr(cls, "_SHARED", None):
            return shared

        else:
            cls._SHARED = cls()  # pyright: ignore[reportConstantRedefinition]
            return cls._SHARED

    def __init__(
        self,
        endpoint: str | None = None,
        api_key: str | None = None,
        timeout: float | None = None,
    ) -> None:
        self._endpoint: str = endpoint or getenv_str(
            "MISTRAL_ENDPOINT",
            default="https://api.mistral.ai",
        )
        self._api_key: str | None = api_key or getenv_str("MISTRAL_API_KEY")
        self._timeout: float = timeout or 60
        self._client: AsyncClient = self._initialize_client()

    def _initialize_client(self) -> AsyncClient:
        return AsyncClient(  # nosec: B113 - false positive
            base_url=self._endpoint,
            headers={
                "Authorization": f"Bearer {self._api_key}",
            },
            timeout=self._timeout,
        )

    @overload
    async def chat_completion(
        self,
        *,
        config: MistralChatConfig,
        messages: list[ChatMessage],
        response_format: dict[str, str] | None,
        tools: list[dict[str, object]] | None = None,
        tool_choice: Literal["auto", "any", "none"] = "auto",
        stream: Literal[True],
    ) -> AsyncIterator[ChatCompletionStreamResponse]: ...

    @overload
    async def chat_completion(
        self,
        *,
        config: MistralChatConfig,
        messages: list[ChatMessage],
        response_format: dict[str, str] | None,
        tools: list[dict[str, object]] | None = None,
        tool_choice: Literal["auto", "any", "none"] = "auto",
    ) -> ChatCompletionResponse: ...

    async def chat_completion(  # noqa: PLR0913
        self,
        *,
        config: MistralChatConfig,
        messages: list[ChatMessage],
        response_format: dict[str, str] | None,
        tools: list[dict[str, object]] | None = None,
        tool_choice: Literal["auto", "any", "none"] = "auto",
        stream: bool = False,
    ) -> AsyncIterator[ChatCompletionStreamResponse] | ChatCompletionResponse:
        if stream:
            raise NotImplementedError("Mistral streaming is not supported yet")

        else:
            if messages[-1]["role"] == "assistant":
                if response_format == {"type": "json_object"}:
                    del messages[-1]  # for json mode ignore prefill

                else:
                    messages[-1]["prefix"] = True  # add prefill parameter indicator

            return await self._create_chat_completion(
                messages=messages,
                model=config.model,
                temperature=config.temperature,
                top_p=config.top_p if not_missing(config.top_p) else None,
                max_tokens=config.max_tokens if not_missing(config.max_tokens) else None,
                response_format=cast(dict[str, str], response_format),
                seed=config.seed if not_missing(config.seed) else None,
                tools=tools,
                tool_choice=tool_choice if tools else None,
                stop=config.stop_sequences if not_missing(config.stop_sequences) else None,
            )

    async def embedding(
        self,
        config: MistralEmbeddingConfig,
        inputs: Sequence[str],
    ) -> list[Sequence[float]]:
        return list(
            chain(
                *await gather(
                    *[
                        self._create_text_embedding(
                            model=config.model,
                            texts=inputs[index : index + config.batch_size],
                        )
                        for index in range(0, len(inputs), config.batch_size)
                    ]
                )
            )
        )

    async def _create_chat_completion(  # noqa: PLR0913
        self,
        model: str,
        temperature: float,
        top_p: float | None,
        seed: int | None,
        max_tokens: int | None,
        response_format: dict[str, str] | None,
        messages: list[ChatMessage],
        tools: list[dict[str, object]] | None,
        tool_choice: str | None,
        stop: Sequence[str] | None,
    ) -> ChatCompletionResponse:
        request_body: dict[str, Any] = {
            "model": model,
            "temperature": temperature,
            "messages": messages,
        }

        if tools:
            request_body["tools"] = tools
        if tool_choice is not None:
            request_body["tool_choice"] = tool_choice
        elif tools:
            request_body["tool_choice"] = "auto"
        if max_tokens:
            request_body["max_tokens"] = max_tokens
        if top_p is not None:
            request_body["top_p"] = top_p
        if seed is not None:
            request_body["random_seed"] = seed
        if stop:
            request_body["stop"] = stop
        if response_format is not None:
            request_body["response_format"] = response_format

        return await self._request(
            model=ChatCompletionResponse,
            method="POST",
            url="v1/chat/completions",
            body=request_body,
        )

    async def _create_text_embedding(
        self,
        model: str,
        texts: Sequence[str],
    ) -> list[Sequence[float]]:
        response: EmbeddingResponse = await self._request(
            model=EmbeddingResponse,
            method="POST",
            url="v1/embeddings",
            body={
                "model": model,
                "input": texts,
            },
        )
        return [element.embedding for element in response.data]

    async def __aenter__(self) -> None:
        if self._client.is_closed:
            self._client = self._initialize_client()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._client.aclose()

    async def _request[Requested: DataModel](  # noqa: PLR0913
        self,
        model: type[Requested],
        method: str,
        url: str,
        query: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        body: DataModel | dict[str, Any] | None = None,
        follow_redirects: bool | None = None,
        timeout: float | None = None,
    ) -> Requested:
        request_headers: dict[str, str]
        if headers:
            request_headers = headers

        else:
            request_headers = {
                "Accept": "application/json",
            }

        body_content: str | None
        match body:
            case None:
                body_content = None

            case body_model if isinstance(body_model, DataModel):
                body_content = body_model.as_json()

            case values:
                body_content = json.dumps(values)

        if body_content:
            request_headers["Content-Type"] = "application/json"

        response: Response
        try:
            response = await self._client.request(
                method=method,
                url=url,
                headers=request_headers,
                params=query,
                content=body_content,
                follow_redirects=follow_redirects or False,
                timeout=timeout,
            )

        except Exception as exc:
            raise MistralException("Network request failed") from exc

        status: HTTPStatus = HTTPStatus(value=response.status_code)
        if status.is_success:
            try:
                return model.from_json(await response.aread())

            except Exception as exc:
                raise MistralException("Failed to decode Mistral response %s", response) from exc

        elif status.is_client_error or status.is_server_error:
            error_body: bytes = await response.aread()
            raise MistralException(
                "Mistral request error: %s %s",
                status,
                error_body.decode("utf-8"),
            )

        else:
            raise MistralException("Network request failed %s", response)



from asyncio import Lock
from collections.abc import Callable, MutableMapping, MutableSequence, Sequence
from typing import Any, Protocol, Self, cast, final, runtime_checkable

from haiway import AttributePath, AttributeRequirement, State

from draive.embedding import Embedded, embed_text, embed_texts
from draive.parameters import DataModel
from draive.similarity import mmr_vector_similarity_search, vector_similarity_search

__all__ = [
    "VectorIndex",
    "VectorIndexing",
    "VectorSearching",
]


@runtime_checkable
class VectorIndexing(Protocol):
    async def __call__[Model: DataModel, Value: str](
        self,
        model: type[Model],
        /,
        values: Sequence[Model],
        indexed_value: Callable[[Model], Value] | AttributePath[Model, Value] | Value,
        **extra: Any,
    ) -> None: ...


@runtime_checkable
class VectorSearching(Protocol):
    async def __call__[Model: DataModel](
        self,
        model: type[Model],
        /,
        query: Sequence[float] | str | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        score_threshold: float | None = None,
        limit: int = 10,
        **extra: Any,
    ) -> Sequence[Model]: ...


@final
class VectorIndex(State):
    @classmethod
    def volatile(cls) -> Self:  # noqa: C901
        lock: Lock = Lock()
        storage: MutableMapping[type[Any], MutableSequence[Embedded[Any]]] = {}

        async def index[Model: DataModel, Value: str](
            model: type[Model],
            /,
            values: Sequence[Model],
            indexed_value: Callable[[Model], Value] | AttributePath[Model, Value] | Value,
            **extra: Any,
        ) -> None:
            async with lock:
                text_selector: Callable[[Model], Value]
                match indexed_value:
                    case Callable() as selector:
                        text_selector = selector

                    case path:
                        assert isinstance(  # nosec: B101
                            path, AttributePath
                        ), "Prepare parameter path by using Self._.path.to.property"
                        text_selector = cast(AttributePath[Model, Value], path).__call__

                embedded_texts: Sequence[Embedded[str]] = await embed_texts(
                    [text_selector(value) for value in values],
                    **extra,
                )

                embedded_models: list[Embedded[Model]] = [
                    Embedded(value=values[index], vector=embedded.vector)
                    for index, embedded in enumerate(embedded_texts)
                ]

                if model in storage:
                    storage[model].extend(embedded_models)

                else:
                    storage[model] = embedded_models

        async def search[Model: DataModel](
            model: type[Model],
            /,
            query: Sequence[float] | str | None = None,
            requirements: AttributeRequirement[Model] | None = None,
            score_threshold: float | None = None,
            limit: int = 10,
            **extra: Any,
        ) -> Sequence[Model]:
            async with lock:
                stored: Sequence[Embedded[Model]] | None = storage.get(model)
                if not stored:
                    return []

                filtered: Sequence[Embedded[Model]]
                if requirements:
                    filtered = [
                        element
                        for element in stored
                        if requirements.check(
                            element.value,
                            raise_exception=False,
                        )
                    ]

                else:
                    filtered = stored

                if not filtered:
                    return []

                query_vector: Sequence[float]
                match query:
                    case None:
                        if limit:
                            return [embedded.value for embedded in filtered[:limit]]

                        else:
                            return [embedded.value for embedded in filtered]

                    case str() as text:
                        embedded_text: Embedded[str] = await embed_text(
                            text,
                            **extra,
                        )
                        query_vector = embedded_text.vector

                    case vector:
                        query_vector = vector

                matching: Sequence[Embedded[Model]] = [
                    filtered[index]
                    for index in vector_similarity_search(
                        query_vector=query_vector,
                        values_vectors=[element.vector for element in filtered],
                        score_threshold=score_threshold,
                        limit=limit * 8,  # feed MMR with more results
                    )
                ]

                if not matching:
                    return []

                return [
                    matching[index].value
                    for index in mmr_vector_similarity_search(
                        query_vector=query_vector,
                        values_vectors=[element.vector for element in matching],
                        limit=limit,
                    )
                ]

        return cls(
            index=index,
            search=search,
        )

    index: VectorIndexing
    search: VectorSearching



from collections.abc import Mapping, Sequence
from itertools import chain
from typing import Self, final, overload

from draive.multimodal.media import MediaContent, MediaKind
from draive.multimodal.text import TextContent
from draive.parameters import DataModel

__all__ = [
    "Multimodal",
    "MultimodalContent",
    "MultimodalContentConvertible",
    "MultimodalContentElement",
]


MultimodalContentElement = TextContent | MediaContent | DataModel
MultimodalContentConvertible = str | MultimodalContentElement


@final
class MultimodalContent(DataModel):
    @classmethod
    def of(
        cls,
        *elements: Self | MultimodalContentConvertible,
        meta: Mapping[str, str | float | int | bool | None] | None = None,
    ) -> Self:
        match elements:
            case [MultimodalContent() as content]:
                # if we got just a single content use it as is
                return content  # pyright: ignore[reportReturnType]

            case elements:
                return cls(
                    parts=tuple(
                        _merge_texts(
                            *chain.from_iterable(
                                _extract_parts(
                                    element,
                                    meta=meta,
                                )
                                for element in elements
                            )
                        )
                    ),
                )

    parts: Sequence[MultimodalContentElement]

    @property
    def has_media(self) -> bool:
        return any(_is_media(part) for part in self.parts)

    @property
    def has_artifacts(self) -> bool:
        return any(_is_artifact(part) for part in self.parts)

    def media(
        self,
        media: MediaKind | None = None,
    ) -> Sequence[MediaContent]:
        if media is None:
            return tuple(part for part in self.parts if isinstance(part, MediaContent))

        else:
            return tuple(
                part for part in self.parts if isinstance(part, MediaContent) and part.kind == media
            )

    def without_media(self) -> Self:
        return self.__class__(
            parts=tuple(part for part in self.parts if not isinstance(part, MediaContent)),
        )

    @overload
    def artifacts(
        self,
        /,
    ) -> Sequence[DataModel]: ...

    @overload
    def artifacts[Artifact: DataModel](
        self,
        model: type[Artifact],
        /,
    ) -> Sequence[Artifact]: ...

    def artifacts[Artifact: DataModel](
        self,
        model: type[Artifact] | None = None,
        /,
    ) -> Sequence[Artifact] | Sequence[DataModel]:
        if model is None:
            return tuple(part for part in self.parts if _is_artifact(part))

        else:
            return tuple(part for part in self.parts if isinstance(part, model))

    def without_artifacts(self) -> Self:
        return self.__class__(
            parts=tuple(part for part in self.parts if not _is_artifact(part)),
        )

    def as_string(
        self,
        joiner: str | None = None,
        include_data: bool = False,
    ) -> str:
        return (joiner if joiner is not None else "").join(
            _as_string(
                element,
                include_data=include_data,
            )
            for element in self.parts
        )

    def appending(
        self,
        *parts: MultimodalContentConvertible,
        merge_text: bool = False,
    ) -> Self:
        if not self.parts:
            if merge_text:
                return self.__class__(
                    parts=tuple(_merge_texts(*(_as_content(element) for element in parts))),
                )

            else:
                return self.__class__(
                    parts=tuple(_as_content(element) for element in parts),
                )

        if merge_text:
            # check the last part
            match self.parts[-1]:
                case TextContent() as text:
                    # if it is a text append merge starting with it
                    return self.__class__(
                        parts=(
                            *self.parts[:-1],
                            *_merge_texts(text, *(_as_content(element) for element in parts)),
                        )
                    )

                case _:
                    # otherwise just append merged items
                    return self.__class__(
                        parts=(
                            *self.parts,
                            *_merge_texts(*(_as_content(element) for element in parts)),
                        )
                    )

        else:
            return self.__class__(
                parts=(
                    *self.parts,
                    *(_as_content(element) for element in parts),
                )
            )

    def extending(
        self,
        *other: Self,
        merge_text: bool = False,
    ) -> Self:
        return self.appending(
            *chain.from_iterable(content.parts for content in other),
            merge_text=merge_text,
        )

    def __bool__(self) -> bool:
        return bool(self.parts) and any(self.parts)

    def __str__(self) -> str:
        return self.as_string()


Multimodal = MultimodalContent | MultimodalContentConvertible


def _extract_parts(  # noqa: PLR0911
    element: Multimodal,
    /,
    meta: Mapping[str, str | float | int | bool | None] | None = None,
) -> Sequence[MultimodalContentElement]:
    match element:
        case MultimodalContent() as content:
            if not content:
                return ()

            elif meta:
                return tuple(
                    _update_meta(
                        meta,
                        element=part,
                    )
                    for part in content.parts
                )

            else:
                return content.parts

        case str() as text:
            if not text:
                return ()

            else:
                return (
                    TextContent(
                        text=text,
                        meta=meta,
                    ),
                )

        case element:
            if not element:
                return ()

            elif meta:
                return (
                    _update_meta(
                        meta,
                        element=element,
                    ),
                )

            else:
                return (element,)


def _as_content(
    element: MultimodalContentConvertible,
    /,
) -> MultimodalContentElement:
    match element:
        case str() as text:
            return TextContent(text=text)

        case element:
            return element


def _is_media(
    element: MultimodalContentElement,
) -> bool:
    return isinstance(
        element,
        MediaContent,
    )


def _is_artifact(
    element: MultimodalContentElement,
) -> bool:
    return not isinstance(
        element,
        TextContent | MediaContent,
    )


def _as_string(
    element: MultimodalContentElement,
    /,
    *,
    include_data: bool,
) -> str:
    match element:
        case TextContent() as text:
            return text.text

        case MediaContent() as media:
            return media.as_string(include_data=include_data)

        case DataModel() as model:
            return str(model)


def _merge_texts(
    *elements: MultimodalContentElement,
) -> Sequence[MultimodalContentElement]:
    if len(elements) <= 1:
        return elements

    result: list[MultimodalContentElement] = []
    last_text_element: TextContent | None = None
    for element in elements:
        match element:
            case TextContent() as text:
                # do not merge texts with different metadata
                if last_text := last_text_element:
                    if last_text.meta == text.meta:
                        last_text_element = TextContent(
                            text=last_text.text + text.text,
                            meta=text.meta,
                        )

                    else:
                        result.append(last_text)
                        last_text_element = text

                else:
                    last_text_element = text

            case other:
                if last_text := last_text_element:
                    result.append(last_text)
                    last_text_element = None

                result.append(other)

    if last_text := last_text_element:
        result.append(last_text)

    return result


def _update_meta(
    meta: Mapping[str, str | float | int | bool | None],
    /,
    element: MultimodalContentElement,
) -> MultimodalContentElement:
    match element:
        case (TextContent() | MediaContent()) as content:
            if current_meta := content.meta:
                return content.updated(meta={**current_meta, **meta})

            else:
                return content.updated(meta=meta)

        case DataModel() as model:
            return model



from draive.prompts.fetch import fetch_prompt, fetch_prompt_list
from draive.prompts.state import PromptRepository
from draive.prompts.template import PromptAvailabilityCheck, PromptTemplate, prompt
from draive.prompts.types import (
    MissingPrompt,
    Prompt,
    PromptDeclaration,
    PromptDeclarationArgument,
    PromptFetching,
    PromptListing,
)

__all__ = [
    "MissingPrompt",
    "Prompt",
    "PromptAvailabilityCheck",
    "PromptDeclaration",
    "PromptDeclarationArgument",
    "PromptFetching",
    "PromptListing",
    "PromptRepository",
    "PromptTemplate",
    "fetch_prompt",
    "fetch_prompt_list",
    "prompt",
]



from collections.abc import AsyncIterator, Iterable, Mapping, Sequence
from typing import (
    Any,
    Literal,
    Protocol,
    Self,
    TypedDict,
    runtime_checkable,
)

from haiway import State

from draive.instructions import Instruction
from draive.multimodal import Multimodal, MultimodalContent
from draive.parameters import DataModel, Field, ParametersSpecification

__all__ = [
    "LMMCompletion",
    "LMMContext",
    "LMMContextElement",
    "LMMInput",
    "LMMInvocating",
    "LMMOutput",
    "LMMStreamChunk",
    "LMMStreamInput",
    "LMMStreamOutput",
    "LMMStreamProperties",
    "LMMStreaming",
    "LMMToolError",
    "LMMToolException",
    "LMMToolFunctionSpecification",
    "LMMToolRequest",
    "LMMToolRequests",
    "LMMToolResponse",
    "LMMToolResponses",
    "LMMToolSelection",
    "LMMToolSpecification",
]


class LMMToolFunctionSpecification(TypedDict):
    name: str
    description: str | None
    parameters: ParametersSpecification


type LMMToolSpecification = LMMToolFunctionSpecification
type LMMOutputSelection = Literal["auto", "text", "image", "audio", "video"] | type[DataModel]
type LMMToolSelection = Literal["auto", "required", "none"] | LMMToolSpecification


class LMMToolException(Exception):
    pass


class LMMToolError(LMMToolException):
    def __init__(
        self,
        *args: object,
        content: MultimodalContent,
    ) -> None:
        super().__init__(*args)
        self.content: MultimodalContent = content


class LMMInput(DataModel):
    @classmethod
    def of(
        cls,
        content: Multimodal,
        /,
    ) -> Self:
        return cls(content=MultimodalContent.of(content))

    content: MultimodalContent

    def __bool__(self) -> bool:
        return bool(self.content)


class LMMCompletion(DataModel):
    @classmethod
    def of(
        cls,
        content: Multimodal,
        /,
    ) -> Self:
        return cls(content=MultimodalContent.of(content))

    content: MultimodalContent

    def __bool__(self) -> bool:
        return bool(self.content)


class LMMToolResponse(DataModel):
    identifier: str
    tool: str
    content: MultimodalContent
    direct: bool
    error: bool


class LMMToolResponses(DataModel):
    responses: Sequence[LMMToolResponse]


class LMMToolRequest(DataModel):
    identifier: str
    tool: str
    arguments: Mapping[str, Any] = Field(default_factory=dict)


class LMMToolRequests(DataModel):
    requests: Sequence[LMMToolRequest]


LMMContextElement = LMMInput | LMMCompletion | LMMToolRequests | LMMToolResponses
LMMContext = Sequence[LMMContextElement]
LMMOutput = LMMCompletion | LMMToolRequests


class LMMStreamChunk(DataModel):
    @classmethod
    def of(
        cls,
        content: Multimodal,
        /,
        eod: bool = False,
    ) -> Self:
        return cls(
            content=MultimodalContent.of(content),
            eod=eod,
        )

    content: MultimodalContent
    eod: bool

    def __bool__(self) -> bool:
        return bool(self.content)


LMMStreamInput = LMMStreamChunk | LMMToolResponse
LMMStreamOutput = LMMStreamChunk | LMMToolRequest


@runtime_checkable
class LMMInvocating(Protocol):
    async def __call__(
        self,
        *,
        instruction: Instruction | str | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        **extra: Any,
    ) -> LMMOutput: ...


class LMMStreamProperties(State):
    instruction: Instruction | str | None = None
    tools: Sequence[LMMToolSpecification] | None


@runtime_checkable
class LMMStreaming(Protocol):
    async def __call__(
        self,
        *,
        properties: AsyncIterator[LMMStreamProperties],
        input: AsyncIterator[LMMStreamInput],  # noqa: A002
        context: LMMContext | None,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput]: ...



from haiway import State

from draive.lmm.types import LMMInvocating, LMMStreaming

__all__ = [
    "LMMInvocating",
    "LMMStream",
]


class LMMInvocation(State):
    invoke: LMMInvocating


class LMMStream(State):
    prepare: LMMStreaming



from haiway import State

from draive.generation.model.default import default_generate_model
from draive.generation.model.types import ModelGenerator

__all__ = [
    "ModelGeneration",
]


class ModelGeneration(State):
    generate: ModelGenerator = default_generate_model



from haiway import State

from draive.steps.default import default_steps_completion
from draive.steps.types import StepsCompleting

__all__: list[str] = [
    "Steps",
]


class Steps(State):
    completion: StepsCompleting = default_steps_completion



from collections.abc import Sequence
from typing import Any, Protocol, runtime_checkable

from haiway import State

from draive.parameters import DataModel

__all__ = [
    "Embedded",
    "ValueEmbedder",
]


class Embedded[Value: DataModel | State | str | bytes](State):
    value: Value
    vector: Sequence[float]


@runtime_checkable
class ValueEmbedder[Value: DataModel | State | str | bytes](Protocol):
    async def __call__(
        self,
        values: Sequence[Value],
        **extra: Any,
    ) -> Sequence[Embedded[Value]]: ...



from asyncio import gather
from collections.abc import Callable, Mapping, Sequence
from typing import Protocol, Self, cast, overload, runtime_checkable

from haiway import AttributePath, ctx, freeze

from draive.evaluation.evaluator import EvaluatorResult, PreparedEvaluator
from draive.parameters import DataModel, Field

__all__ = [
    "EvaluationScenarioResult",
    "ScenarioEvaluator",
    "ScenarioEvaluatorDefinition",
    "ScenarioEvaluatorResult",
    "evaluation_scenario",
]


class ScenarioEvaluatorResult(DataModel):
    name: str = Field(
        description="Name of evaluated scenario",
    )
    evaluations: Sequence[EvaluatorResult] = Field(
        description="Scenario evaluation results",
    )
    meta: Mapping[str, str | float | int | bool | None] | None = Field(
        description="Additional evaluation metadata",
        default=None,
    )

    @property
    def passed(self) -> bool:
        # empty evaluations is equivalent of failure
        return len(self.evaluations) > 0 and all(case.passed for case in self.evaluations)

    def report(self) -> str:
        report: str = "\n- ".join(
            result.report() for result in self.evaluations if not result.passed
        )

        if report:  # nonempty report contains failing reports
            meta_values: str = (
                f"\n{'\n'.join(f'{key}: {value}' for key, value in self.meta.items())}"
                if self.meta
                else "N/A"
            )
            return f"Scenario {self.name}, meta: {meta_values}\n---\n{report}"

        elif not self.evaluations:
            return f"Scenario {self.name} empty!"

        else:
            return f"Scenario {self.name} passed!"


class EvaluationScenarioResult(DataModel):
    @classmethod
    async def evaluating[Value](
        cls,
        value: Value,
        /,
        evaluators: PreparedEvaluator[Value],
        *_evaluators: PreparedEvaluator[Value],
        meta: Mapping[str, str | float | int | bool | None] | None = None,
    ) -> Self:
        return cls(
            evaluations=tuple(
                await gather(
                    *[evaluator(value) for evaluator in [evaluators, *_evaluators]],
                    return_exceptions=False,
                ),
            ),
            meta=meta,
        )

    evaluations: Sequence[EvaluatorResult] = Field(
        description="Scenario evaluation results",
    )
    meta: Mapping[str, str | float | int | bool | None] | None = Field(
        description="Additional evaluation metadata",
        default=None,
    )


@runtime_checkable
class PreparedScenarioEvaluator[Value](Protocol):
    async def __call__(
        self,
        value: Value,
        /,
    ) -> ScenarioEvaluatorResult: ...


@runtime_checkable
class ScenarioEvaluatorDefinition[Value, **Args](Protocol):
    @property
    def __name__(self) -> str: ...

    async def __call__(
        self,
        value: Value,
        /,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Sequence[EvaluatorResult] | EvaluationScenarioResult: ...


class ScenarioEvaluator[Value, **Args]:
    def __init__(
        self,
        name: str,
        definition: ScenarioEvaluatorDefinition[Value, Args],
        meta: Mapping[str, str | float | int | bool | None] | None = None,
    ) -> None:
        self.name: str = name
        self._definition: ScenarioEvaluatorDefinition[Value, Args] = definition
        self.meta: Mapping[str, str | float | int | bool | None] | None = meta

        freeze(self)

    def prepared(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> PreparedScenarioEvaluator[Value]:
        async def evaluate(
            value: Value,
        ) -> ScenarioEvaluatorResult:
            return await self(
                value,
                *args,
                **kwargs,
            )

        return evaluate

    def with_meta(
        self,
        meta: Mapping[str, str | float | int | bool | None],
        /,
    ) -> Self:
        return self.__class__(
            name=self.name,
            definition=self._definition,
            meta={**self.meta, **meta} if self.meta else meta,
        )

    def contra_map[Mapped](
        self,
        mapping: Callable[[Mapped], Value] | AttributePath[Mapped, Value] | Value,
        /,
    ) -> "ScenarioEvaluator[Mapped, Args]":
        mapper: Callable[[Mapped], Value]
        match mapping:
            case Callable() as function:  # pyright: ignore[reportUnknownVariableType]
                mapper = function

            case path:
                assert isinstance(  # nosec: B101
                    path, AttributePath
                ), "Prepare parameter path by using Self._.path.to.property"
                mapper = cast(AttributePath[Mapped, Value], path).__call__

        async def evaluation(
            value: Mapped,
            /,
            *args: Args.args,
            **kwargs: Args.kwargs,
        ) -> Sequence[EvaluatorResult] | EvaluationScenarioResult:
            return await self._definition(
                mapper(value),
                *args,
                **kwargs,
            )

        return ScenarioEvaluator[Mapped, Args](
            name=self.name,
            definition=evaluation,
        )

    async def __call__(
        self,
        value: Value,
        /,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> ScenarioEvaluatorResult:
        try:
            match await self._definition(
                value,
                *args,
                **kwargs,
            ):
                case EvaluationScenarioResult() as result:
                    meta: Mapping[str, str | float | int | bool | None] | None
                    if self.meta:
                        if result.meta:
                            meta = {**self.meta, **result.meta}

                        else:
                            meta = self.meta

                    elif result.meta:
                        meta = result.meta

                    else:
                        meta = None

                    return ScenarioEvaluatorResult(
                        name=self.name,
                        evaluations=result.evaluations,
                        meta=meta,
                    )

                case [*results]:
                    return ScenarioEvaluatorResult(
                        name=self.name,
                        evaluations=tuple(results),
                        meta=self.meta,
                    )

        except Exception as exc:
            ctx.log_error(
                f"Scenario evaluator `{self.name}` failed, using empty fallback result",
                exception=exc,
            )

            return ScenarioEvaluatorResult(
                name=self.name,
                evaluations=(),
                meta={**(self.meta or {}), "exception": str(exc)},
            )


@overload
def evaluation_scenario[Value, **Args](
    definition: ScenarioEvaluatorDefinition[Value, Args],
    /,
) -> ScenarioEvaluator[Value, Args]: ...


@overload
def evaluation_scenario[Value, **Args](
    *,
    name: str,
) -> Callable[
    [ScenarioEvaluatorDefinition[Value, Args]],
    ScenarioEvaluator[Value, Args],
]: ...


def evaluation_scenario[Value, **Args](  # pyright: ignore[reportInconsistentOverload] - this seems to be pyright false positive/error
    definition: ScenarioEvaluatorDefinition[Value, Args] | None = None,
    /,
    *,
    name: str | None = None,
) -> (
    Callable[
        [ScenarioEvaluatorDefinition[Value, Args]],
        ScenarioEvaluator[Value, Args],
    ]
    | ScenarioEvaluator[Value, Args]
):
    def wrap(
        definition: ScenarioEvaluatorDefinition[Value, Args],
    ) -> ScenarioEvaluator[Value, Args]:
        return ScenarioEvaluator(
            name=name or definition.__name__,
            definition=definition,
        )

    if definition:
        return wrap(definition)

    else:
        return wrap



from draive.generation.model.call import generate_model
from draive.generation.model.state import ModelGeneration
from draive.generation.model.types import ModelGenerator, ModelGeneratorDecoder

__all__ = [
    "ModelGeneration",
    "ModelGenerator",
    "ModelGeneratorDecoder",
    "generate_model",
]



from types import TracebackType
from typing import final

__all__ = [
    "VolatileMemory",
]


@final
class VolatileMemory[Item]:
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        pass  # noop



from collections.abc import Sequence
from typing import Any

from fastembed.text.text_embedding import (  # pyright: ignore[reportMissingTypeStubs]
    TextEmbedding as FastembedTextEmbedding,
)
from haiway import asynchronous, ctx

from draive.embedding import Embedded, TextEmbedding

__all__ = [
    "fastembed_text_embedding",
]


async def fastembed_text_embedding(
    model_name: str,
    cache_dir: str | None = "./embedding_models/",
) -> TextEmbedding:
    embedding_model: FastembedTextEmbedding = await _text_embedding_model(
        model_name=model_name,
        cache_dir=cache_dir,
    )

    async def fastembed_embed_text(
        values: Sequence[str],
        **extra: Any,
    ) -> Sequence[Embedded[str]]:
        with ctx.scope("text_embedding"):
            return await _fastembed_text_embedding(
                embedding_model,
                values,
            )

    return TextEmbedding(embed=fastembed_embed_text)


@asynchronous
def _text_embedding_model(
    model_name: str,
    cache_dir: str | None = "./embedding_models/",
) -> FastembedTextEmbedding:
    return FastembedTextEmbedding(
        model_name=model_name,
        cache_dir=cache_dir,
    )


@asynchronous
def _fastembed_text_embedding(
    embedding_model: FastembedTextEmbedding,
    texts: Sequence[str],
    /,
) -> Sequence[Embedded[str]]:
    return [
        Embedded(
            value=value,
            vector=embedding.tolist(),
        )
        for (
            value,
            embedding,  # pyright: ignore[reportUnknownVariableType]
        ) in zip(
            texts,
            embedding_model.embed(  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType, reportUnknownArgumentType]
                texts
            ),
            strict=True,
        )
    ]



from collections.abc import Mapping, Sequence
from typing import Any, Literal, overload

from haiway import ctx

from draive.resources.state import ResourceRepository
from draive.resources.types import MissingResource, Resource, ResourceDeclaration

__all__ = [
    "fetch_resource",
    "fetch_resource_list",
]


async def fetch_resource_list(
    **extra: Any,
) -> Sequence[ResourceDeclaration]:
    return await ctx.state(ResourceRepository).list(**extra)


@overload
async def fetch_resource(
    reference: ResourceDeclaration | str,
    /,
    *,
    default: Resource | None = None,
    variables: Mapping[str, str] | None = None,
    **extra: Any,
) -> Resource | None: ...


@overload
async def fetch_resource(
    reference: ResourceDeclaration | str,
    /,
    *,
    default: Resource,
    **extra: Any,
) -> Resource: ...


@overload
async def fetch_resource(
    reference: ResourceDeclaration | str,
    /,
    *,
    default: Resource | None = None,
    required: Literal[True],
    **extra: Any,
) -> Resource: ...


async def fetch_resource(
    reference: ResourceDeclaration | str,
    /,
    *,
    default: Resource | None = None,
    required: bool = True,
    **extra: Any,
) -> Resource | None:
    uri: str = reference if isinstance(reference, str) else reference.uri

    match await ctx.state(ResourceRepository).fetch(
        uri,
        **extra,
    ):
        case None:
            match default:
                case None:
                    if required:
                        raise MissingResource(f"Missing resource: '{uri}'")

                    else:
                        return None

                case Resource() as resource:
                    return resource

        case resource:
            return resource



from draive.utils.memory import Memory
from draive.utils.rate_limit import RateLimitError
from draive.utils.splitting import split_sequence
from draive.utils.streams import AsyncStream, ConstantStream, FixedStream

__all__ = [
    "AsyncStream",
    "ConstantStream",
    "FixedStream",
    "Memory",
    "RateLimitError",
    "split_sequence",
]



from haiway import State, async_always

from draive.instructions.types import InstructionFetching

__all__ = [
    "InstructionsRepository",
]


class InstructionsRepository(State):
    fetch: InstructionFetching = async_always(None)



from draive.instructions.fetch import fetch_instruction
from draive.instructions.file import instructions_file
from draive.instructions.state import InstructionsRepository
from draive.instructions.template import InstructionTemplate, instruction
from draive.instructions.types import Instruction, InstructionFetching, MissingInstruction

__all__ = [
    "Instruction",
    "InstructionFetching",
    "InstructionTemplate",
    "InstructionsRepository",
    "MissingInstruction",
    "fetch_instruction",
    "instruction",
    "instructions_file",
]



import json
from base64 import b64encode
from collections.abc import AsyncGenerator, AsyncIterator, Iterable, Sequence
from itertools import chain
from typing import Any, Literal, cast
from uuid import uuid4

from haiway import ArgumentsTrace, ResultTrace, ctx, not_missing
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionContentPartParam,
    ChatCompletionMessage,
    ChatCompletionMessageParam,
)
from openai.types.chat.chat_completion_chunk import Choice, ChoiceDeltaToolCall

from draive.instructions import Instruction
from draive.lmm import (
    LMMCompletion,
    LMMContext,
    LMMContextElement,
    LMMInput,
    LMMInvocation,
    LMMOutput,
    LMMOutputSelection,
    LMMStream,
    LMMStreamChunk,
    LMMStreamInput,
    LMMStreamOutput,
    LMMStreamProperties,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.lmm.types import LMMToolResponses
from draive.metrics import TokenUsage
from draive.multimodal import (
    MediaContent,
    MultimodalContent,
    MultimodalContentElement,
    TextContent,
)
from draive.openai.client import OpenAIClient
from draive.openai.config import OpenAIChatConfig, OpenAISystemFingerprint
from draive.openai.types import OpenAIException
from draive.parameters import DataModel

__all__ = [
    "openai_lmm",
    "openai_streaming_lmm",
]


def openai_lmm(
    client: OpenAIClient | None = None,
    /,
) -> LMMInvocation:
    client = client or OpenAIClient.shared()

    async def lmm_invocation(
        *,
        instruction: Instruction | str | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        **extra: Any,
    ) -> LMMOutput:
        with ctx.scope("openai_lmm_invocation"):
            ctx.record(
                ArgumentsTrace.of(
                    instruction=instruction,
                    context=context,
                    tool_selection=tool_selection,
                    tools=tools,
                    output=output,
                    **extra,
                ),
            )
            config: OpenAIChatConfig = ctx.state(OpenAIChatConfig).updated(**extra)
            ctx.record(config)

            messages: list[ChatCompletionMessageParam] = list(
                chain.from_iterable(
                    [
                        _convert_context_element(config=config, element=element)
                        for element in context
                    ]
                )
            )

            if instruction:
                messages = [
                    {
                        "role": "system",
                        "content": Instruction.of(instruction).format(),
                    },
                    *messages,
                ]

            return await _chat_completion(
                client=client,
                config=config,
                output=output,
                messages=messages,
                tools=tools,
                tool_selection=tool_selection,
            )

    return LMMInvocation(invoke=lmm_invocation)


def openai_streaming_lmm(
    client: OpenAIClient | None = None,
    /,
) -> LMMStream:
    client = client or OpenAIClient.shared()

    async def lmm_stream(
        *,
        properties: AsyncIterator[LMMStreamProperties],
        input: AsyncIterator[LMMStreamInput],  # noqa: A002
        context: LMMContext | None,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput]:
        config: OpenAIChatConfig = ctx.state(OpenAIChatConfig).updated(**extra)
        ctx.record(config)

        context_elements: Sequence[LMMContextElement]
        match context:
            case None:
                context_elements = ()

            case [*elements]:
                context_elements = elements

        return _chat_stream(
            client=client,
            config=config,
            properties=properties,
            input=input,
            context=list(
                chain.from_iterable(
                    [
                        _convert_context_element(config=config, element=element)
                        for element in context_elements
                    ]
                )
            ),
        )

    return LMMStream(prepare=lmm_stream)


def _convert_content_element(
    element: MultimodalContentElement,
    config: OpenAIChatConfig,
) -> ChatCompletionContentPartParam:
    match element:
        case TextContent() as text:
            return {
                "type": "text",
                "text": text.text,
            }

        case MediaContent() as media:
            if media.kind != "image":
                raise ValueError("Unsupported message content", media)

            url: str
            match media.source:
                case str() as string:
                    url = string

                case bytes() as data:
                    url = f"data:{media.media};base64,{b64encode(data).decode()}"

            return {
                "type": "image_url",
                "image_url": {
                    "url": url,
                    "detail": cast(Literal["auto", "low", "high"], config.vision_details)
                    if not_missing(config.vision_details)
                    else "auto",
                },
            }

        case DataModel() as data:
            return {
                "type": "text",
                "text": data.as_json(),
            }


def _convert_context_element(
    element: LMMContextElement,
    config: OpenAIChatConfig,
) -> Iterable[ChatCompletionMessageParam]:
    match element:
        case LMMInput() as input:
            return (
                {
                    "role": "user",
                    "content": [
                        _convert_content_element(
                            element=element,
                            config=config,
                        )
                        for element in input.content.parts
                    ],
                },
            )

        case LMMCompletion() as completion:
            # TODO: OpenAI models generating media?
            return (
                {
                    "role": "assistant",
                    "content": completion.content.as_string(),
                },
            )

        case LMMToolRequests() as tool_requests:
            return (
                {
                    "role": "assistant",
                    "tool_calls": [
                        {
                            "id": request.identifier,
                            "type": "function",
                            "function": {
                                "name": request.tool,
                                "arguments": json.dumps(dict(request.arguments)),
                            },
                        }
                        for request in tool_requests.requests
                    ],
                },
            )

        case LMMToolResponses() as tool_responses:
            return (
                {
                    "role": "tool",
                    "tool_call_id": response.identifier,
                    "content": response.content.as_string(),
                }
                for response in tool_responses.responses
            )


async def _chat_completion(  # noqa: C901, PLR0912, PLR0913
    *,
    client: OpenAIClient,
    config: OpenAIChatConfig,
    output: LMMOutputSelection,
    messages: list[ChatCompletionMessageParam],
    tools: Iterable[LMMToolSpecification] | None,
    tool_selection: LMMToolSelection,
) -> LMMOutput:
    completion: ChatCompletion
    match tool_selection:
        case "auto":
            completion = await client.chat_completion(
                config=config,
                messages=messages,
                response_format=output,
                tools=[
                    {
                        "type": "function",
                        "function": {
                            "name": tool["name"],
                            "description": tool["description"] or "",
                            "parameters": dict[str, object](tool["parameters"]),
                        },
                    }
                    for tool in tools
                ]
                if tools
                else None,
                tool_choice="auto",
            )

        case "none":
            completion = await client.chat_completion(
                config=config,
                messages=messages,
                response_format=output,
                tools=None,
                tool_choice="none",
            )

        case "required":
            completion = await client.chat_completion(
                config=config,
                messages=messages,
                response_format=output,
                tools=[
                    {
                        "type": "function",
                        "function": {
                            "name": tool["name"],
                            "description": tool["description"] or "",
                            "parameters": dict[str, object](tool["parameters"]),
                        },
                    }
                    for tool in tools
                ]
                if tools
                else None,
                tool_choice="required",
            )

        case tool:
            completion = await client.chat_completion(
                config=config,
                messages=messages,
                response_format=output,
                tools=[
                    {
                        "type": "function",
                        "function": {
                            "name": tool["name"],
                            "description": tool["description"] or "",
                            "parameters": dict[str, object](tool["parameters"]),
                        },
                    }
                    for tool in tools
                ]
                if tools
                else None,
                tool_choice={
                    "type": "function",
                    "function": {
                        "name": tool["name"],
                    },
                },
            )

    if usage := completion.usage:
        ctx.record(
            TokenUsage.for_model(
                config.model,
                input_tokens=usage.prompt_tokens,
                output_tokens=usage.completion_tokens,
            ),
        )

    if not completion.choices:
        raise OpenAIException("Invalid OpenAI completion - missing messages!", completion)

    if fingerprint := completion.system_fingerprint:
        ctx.record(OpenAISystemFingerprint(system_fingerprint=fingerprint))

    completion_message: ChatCompletionMessage = completion.choices[0].message
    match completion.choices[0].finish_reason:
        case "tool_calls":
            if (tool_calls := completion_message.tool_calls) and (tools := tools):
                ctx.record(ResultTrace.of(tool_calls))
                return LMMToolRequests(
                    requests=[
                        LMMToolRequest(
                            identifier=call.id,
                            tool=call.function.name,
                            arguments=json.loads(call.function.arguments),
                        )
                        for call in tool_calls
                    ]
                )

            else:
                raise OpenAIException("Invalid OpenAI completion", completion)

        case "stop":
            if (tool_calls := completion_message.tool_calls) and (tools := tools):
                ctx.record(ResultTrace.of(tool_calls))
                return LMMToolRequests(
                    requests=[
                        LMMToolRequest(
                            identifier=call.id,
                            tool=call.function.name,
                            arguments=json.loads(call.function.arguments),
                        )
                        for call in tool_calls
                    ]
                )

            elif content := completion_message.content:
                ctx.record(ResultTrace.of(content))
                return LMMCompletion.of(content)

            else:
                raise OpenAIException("Invalid OpenAI completion", completion)

        case other:
            raise OpenAIException(f"Unexpected finish reason: {other}")


async def _chat_stream(  # noqa: C901, PLR0912, PLR0915
    *,
    client: OpenAIClient,
    config: OpenAIChatConfig,
    properties: AsyncIterator[LMMStreamProperties],
    input: AsyncIterator[LMMStreamInput],  # noqa: A002
    context: list[ChatCompletionMessageParam],
) -> AsyncGenerator[LMMStreamOutput, None]:
    # track requested tool calls out of the loop
    pending_tool_calls: set[str] = set()
    # before each call check for updated properties - this supposed to be an infinite loop
    async for current_properties in properties:
        # for each call accumulate input first
        input_buffer: MultimodalContent = MultimodalContent.of()
        async for chunk in input:
            match chunk:
                # gether input content chunks until marked as end
                case LMMStreamChunk() as content_chunk:
                    input_buffer = input_buffer.appending(content_chunk.content)
                    if content_chunk.eod:
                        context.append(
                            {
                                "role": "user",
                                "content": [
                                    _convert_content_element(
                                        element=element,
                                        config=config,
                                    )
                                    for element in input_buffer.parts
                                ],
                            }
                        )
                        break  # we are supporting only completed input messages with this api

                # accumulate tool results directly in context
                case tool_result:
                    pending_tool_calls.remove(tool_result.identifier)
                    context.append(
                        {
                            "role": "tool",
                            "tool_call_id": tool_result.identifier,
                            "content": tool_result.content.as_string(),
                        }
                    )
                    # when there is no pending input and we got all requested tool results
                    if not input_buffer and not pending_tool_calls:
                        break  # then we can request completion again

        else:  # finalize streaming if input is finished
            return

        # prepare context for the next request by extending with instructions if any
        request_context: list[ChatCompletionMessageParam]
        if instruction := current_properties.instruction:
            request_context = [
                {
                    "role": "system",
                    "content": Instruction.of(instruction).format(),
                },
                *context,
            ]

        else:
            request_context = context

        accumulated_result: MultimodalContent = MultimodalContent.of()
        accumulated_tool_calls: list[ChoiceDeltaToolCall] = []
        async for part in await client.chat_completion(
            config=config,
            messages=request_context,
            tools=[
                {
                    "type": "function",
                    "function": {
                        "name": tool["name"],
                        "description": tool["description"] or "",
                        "parameters": dict[str, object](tool["parameters"]),
                    },
                }
                for tool in current_properties.tools
            ]
            if current_properties.tools
            else None,
            tool_choice="auto" if current_properties.tools else "none",
            stream=True,
        ):
            if part.choices:  # usage part does not contain choices
                # we are always requesting single result - no need to take care of indices
                element: Choice = part.choices[0]
                # get the tool calls parts first
                if tool_calls := element.delta.tool_calls:
                    # tool calls come in parts, we have to merge them manually
                    for call in tool_calls:
                        try:
                            tool_call: ChoiceDeltaToolCall = next(
                                tool_call
                                for tool_call in accumulated_tool_calls
                                if tool_call.index == call.index
                            )

                            if call.id:
                                if tool_call.id is not None:
                                    tool_call.id += call.id
                                else:
                                    tool_call.id = call.id
                            else:
                                pass

                            if call.function is None:
                                continue

                            if tool_call.function is None:
                                tool_call.function = call.function
                                continue

                            if call.function.name:
                                if tool_call.function.name is not None:
                                    tool_call.function.name += call.function.name
                                else:
                                    tool_call.function.name = call.function.name
                            else:
                                pass

                            if call.function.arguments:
                                if tool_call.function.arguments is not None:
                                    tool_call.function.arguments += call.function.arguments
                                else:
                                    tool_call.function.arguments = call.function.arguments
                            else:
                                pass

                        except (StopIteration, StopAsyncIteration):
                            accumulated_tool_calls.append(call)

                # then process content
                if element.delta.content is not None:
                    content_chunk: LMMStreamChunk = LMMStreamChunk.of(element.delta.content)
                    accumulated_result = accumulated_result.appending(
                        content_chunk.content,
                        merge_text=True,
                    )
                    yield content_chunk

                if finish_reason := element.finish_reason:
                    match finish_reason:
                        case "stop":
                            # send completion chunk - openAI sends it without an actual content
                            yield LMMStreamChunk.of(
                                MultimodalContent.of(),
                                eod=True,
                            )

                        case "tool_calls":
                            if accumulated_tool_calls:
                                context.append(
                                    {
                                        "role": "assistant",
                                        "tool_calls": [
                                            {
                                                "id": request.id,
                                                "type": "function",
                                                "function": {
                                                    "name": request.function.name,
                                                    "arguments": request.function.arguments or "{}",
                                                },
                                            }
                                            for request in accumulated_tool_calls
                                            if request.id
                                            and request.function
                                            and request.function.name
                                        ],
                                    }
                                )
                            if accumulated_result:
                                context.append(
                                    {
                                        "role": "assistant",
                                        "content": accumulated_result.as_string(),
                                    }
                                )

                            for call in accumulated_tool_calls:
                                if not call.function:
                                    continue  # skip partial calls
                                if not call.function.name:
                                    continue  # skip calls with missing names

                                call_identifier: str = call.id or uuid4().hex
                                pending_tool_calls.add(call_identifier)
                                # send tool requests when ensured that all were completed
                                yield LMMToolRequest(
                                    identifier=call_identifier,
                                    tool=call.function.name,
                                    arguments=json.loads(call.function.arguments)
                                    if call.function.arguments
                                    else {},
                                )

                        case other:
                            raise OpenAIException(f"Unexpected finish reason: {other}")

            elif usage := part.usage:  # record usage if able (expected in the last part)
                ctx.record(
                    TokenUsage.for_model(
                        config.model,
                        input_tokens=usage.prompt_tokens,
                        output_tokens=usage.completion_tokens,
                    ),
                )

                if fingerprint := part.system_fingerprint:
                    ctx.record(OpenAISystemFingerprint(system_fingerprint=fingerprint))

            else:
                ctx.log_warning("Unexpected OpenAI streaming part: %s", part)



from haiway import State

from draive.embedding.types import ValueEmbedder

__all__ = [
    "ImageEmbedding",
    "TextEmbedding",
]


class TextEmbedding(State):
    embed: ValueEmbedder[str]


class ImageEmbedding(State):
    embed: ValueEmbedder[bytes]



from collections.abc import Sequence
from io import BytesIO
from typing import Any

from fastembed.image.image_embedding import (  # pyright: ignore[reportMissingTypeStubs]
    ImageEmbedding,
)
from haiway import asynchronous, ctx

from draive.embedding import Embedded, ValueEmbedder

__all__ = [
    "fastembed_image_embedding",
]


async def fastembed_image_embedding(
    model_name: str,
    cache_dir: str | None = "./embedding_models/",
    /,
) -> ValueEmbedder[bytes]:
    embedding_model: ImageEmbedding = await _image_embedding_model(
        model_name=model_name,
        cache_dir=cache_dir,
    )

    async def fastembed_embed_image(
        values: Sequence[bytes],
        **extra: Any,
    ) -> Sequence[Embedded[bytes]]:
        with ctx.scope("image_embedding"):
            return await _fastembed_image_embedding(
                embedding_model,
                values,
            )

    return fastembed_embed_image


@asynchronous
def _image_embedding_model(
    model_name: str,
    cache_dir: str | None,
) -> ImageEmbedding:
    return ImageEmbedding(
        model_name=model_name,
        cache_dir=cache_dir,
    )


@asynchronous
def _fastembed_image_embedding(
    embedding_model: ImageEmbedding,
    images: Sequence[bytes],
    /,
) -> Sequence[Embedded[bytes]]:
    return [
        Embedded(
            value=value,  # pyright: ignore[reportUnknownVariableType, reportUnknownArgumentType]
            vector=embedding.tolist(),  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
        )
        for (
            value,  # pyright: ignore[reportUnknownVariableType]
            embedding,  # pyright: ignore[reportUnknownVariableType]
        ) in zip(
            images,
            embedding_model.embed(  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType, reportUnknownArgumentType]
                [BytesIO(image) for image in images]  # pyright: ignore[reportArgumentType]
            ),
            strict=True,
        )
    ]



from draive.resources.fetch import fetch_resource, fetch_resource_list
from draive.resources.state import ResourceRepository
from draive.resources.template import ResourceTemplate, resource
from draive.resources.types import (
    MissingResource,
    Resource,
    ResourceContent,
    ResourceDeclaration,
    ResourceFetching,
    ResourceListing,
)

__all__ = [
    "MissingResource",
    "Resource",
    "ResourceContent",
    "ResourceDeclaration",
    "ResourceFetching",
    "ResourceListing",
    "ResourceRepository",
    "ResourceTemplate",
    "fetch_resource",
    "fetch_resource_list",
    "resource",
]



from collections.abc import Iterable
from typing import Any, Literal, cast

from haiway import ArgumentsTrace, ResultTrace, ctx

from draive.bedrock.client import BedrockClient
from draive.bedrock.config import BedrockChatConfig
from draive.bedrock.models import ChatCompletionResponse, ChatMessage, ChatMessageContent, ChatTool
from draive.bedrock.types import BedrockException
from draive.instructions import Instruction
from draive.lmm import (
    LMMCompletion,
    LMMContext,
    LMMContextElement,
    LMMInput,
    LMMInvocation,
    LMMOutput,
    LMMOutputSelection,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponses,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.metrics.tokens import TokenUsage
from draive.multimodal import (
    MediaContent,
    MultimodalContent,
    MultimodalContentElement,
    TextContent,
)
from draive.parameters import DataModel

__all__ = [
    "bedrock_lmm",
]


def bedrock_lmm(
    client: BedrockClient | None = None,
    /,
) -> LMMInvocation:
    client = client or BedrockClient.shared()

    async def lmm_invocation(
        *,
        instruction: Instruction | str | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        **extra: Any,
    ) -> LMMOutput:
        with ctx.scope("bedrock_lmm_invocation"):
            ctx.record(
                ArgumentsTrace.of(
                    instruction=instruction,
                    context=context,
                    tools=tools,
                    tool_selection=tool_selection,
                    output=output,
                    **extra,
                ),
            )
            config: BedrockChatConfig = ctx.state(BedrockChatConfig).updated(**extra)
            ctx.record(config)

            match output:
                case "auto" | "text":
                    pass

                case "image":
                    raise NotImplementedError("image output is not supported by bedrock")

                case "audio":
                    raise NotImplementedError("audio output is not supported by bedrock")

                case "video":
                    raise NotImplementedError("video output is not supported by bedrock")

                case _:
                    raise NotImplementedError("model output is not supported by bedrock")



            messages: list[ChatMessage] = [
                _convert_context_element(element) for element in context
            ]

            tools_list: list[ChatTool] = [_convert_tool(tool) for tool in tools or []]
            require_tool: bool
            match tool_selection:
                case "required":
                    require_tool = True

                case _:
                    require_tool = False

            return await _chat_completion(
                client=client,
                config=config,
                instruction=Instruction.formatted(instruction),
                messages=messages,
                tools_list=tools_list,
                require_tool=require_tool,
            )

    return LMMInvocation(invoke=lmm_invocation)


def _convert_content_element(
    element: MultimodalContentElement,
) -> ChatMessageContent:
    match element:
        case TextContent() as text:
            return {"text": text.text}

        case MediaContent() as media:
            if media.kind != "image" or isinstance(media.source, str):
                raise ValueError("Unsupported message content", media)

            image_format: Literal["png", "jpeg", "gif"]
            match media.media:
                case "image/png":
                    image_format = "png"

                case "image/jpeg":
                    image_format = "jpeg"

                case "image/gif":
                    image_format = "gif"

                case _:
                    raise ValueError("Unsupported message content", media)

            return {
                "image": {
                    "format": image_format,
                    "source": {"bytes": media.source},
                }
            }

        case DataModel() as data:
            return {"text": data.as_json()}


def _convert_context_element(
    element: LMMContextElement,
) -> ChatMessage:
    match element:
        case LMMInput() as input:
            return ChatMessage(
                role="user",
                content=[_convert_content_element(part) for part in input.content.parts],
            )

        case LMMCompletion() as completion:
            return ChatMessage(
                role="assistant",
                content=[_convert_content_element(part) for part in completion.content.parts],
            )

        case LMMToolRequests() as requests:
            return ChatMessage(
                role="assistant",
                content=[
                    {
                        "toolUse": {
                            "toolUseId": request.identifier,
                            "name": request.tool,
                            "input": request.arguments,
                        }
                    }
                    for request in requests.requests
                ],
            )

        case LMMToolResponses() as tool_responses:
            return ChatMessage(
                role="user",
                content=[
                    {
                        "toolResult": {
                            "toolUseId": response.identifier,
                            "content": [
                                cast(Any, _convert_content_element(part))
                                for part in response.content.parts
                            ],
                            "status": "error" if response.error else "success",
                        },
                    }
                    for response in tool_responses.responses
                ],
            )


def _convert_tool(tool: LMMToolSpecification) -> ChatTool:
    return {
        "name": tool["name"],
        "description": tool["description"] or "",
        "inputSchema": {"json": tool["parameters"]},
    }


async def _chat_completion(  # noqa: PLR0913
    *,
    client: BedrockClient,
    config: BedrockChatConfig,
    instruction: str | None,
    messages: list[ChatMessage],
    tools_list: list[ChatTool],
    require_tool: bool,
) -> LMMOutput:
    completion: ChatCompletionResponse = await client.chat_completion(
        config=config,
        instruction=instruction,
        messages=messages,
        tools=tools_list,
        require_tool=require_tool,
    )

    ctx.record(
        TokenUsage.for_model(
            config.model,
            input_tokens=completion["usage"]["inputTokens"],
            output_tokens=completion["usage"]["outputTokens"],
        ),
    )

    message_parts: list[MultimodalContentElement] = []
    tool_calls: list[LMMToolRequest] = []
    for part in completion["output"]["message"]["content"]:
        match part:
            case {"text": str() as text}:
                message_parts.append(TextContent(text=text))

            case {"image": {"format": str() as data_format, "source": {"bytes": bytes() as data}}}:
                media_type: Any
                match data_format:
                    case "png":
                        media_type = "image/png"

                    case "jpeg":
                        media_type = "image/jpeg"

                    case "gif":
                        media_type = "image/gif"

                    case _:  # pyright: ignore[reportUnnecessaryComparison]
                        media_type = "image"

                message_parts.append(
                    MediaContent.data(
                        data,
                        media=media_type,
                    )
                )

            case {
                "toolUse": {
                    "toolUseId": str() as identifier,
                    "name": str() as tool,
                    "input": arguments,
                }
            }:
                tool_calls.append(
                    LMMToolRequest(
                        identifier=identifier,
                        tool=tool,
                        arguments=arguments,
                    )
                )

            case _:
                pass

    match completion["stopReason"]:
        case "end_turn" | "stop_sequence":
            message_completion = LMMCompletion(content=MultimodalContent.of(*message_parts))
            ctx.record(ResultTrace.of(message_completion))
            return message_completion

        case "tool_use":
            tools_completion = LMMToolRequests(requests=tool_calls)
            ctx.record(ResultTrace.of(tools_completion))
            return tools_completion

        case _:
            raise BedrockException("Invalid Bedrock response")



from draive.tools.tool import AnyTool, Tool, ToolAvailabilityCheck, tool
from draive.tools.toolbox import ExternalToolbox, Toolbox, ToolboxFetching

__all__ = [
    "AnyTool",
    "ExternalToolbox",
    "Tool",
    "ToolAvailabilityCheck",
    "Toolbox",
    "ToolboxFetching",
    "tool",
]



from collections.abc import Callable, Mapping, Sequence
from datetime import date, datetime, time
from enum import IntEnum, StrEnum
from types import EllipsisType, NoneType, UnionType
from typing import Any, Literal, NotRequired, Required, TypedDict, Union, cast, final, is_typeddict
from typing import Mapping as MappingType  # noqa: UP035
from typing import Sequence as SequenceType  # noqa: UP035
from uuid import UUID

from haiway import Missing
from haiway.state import AttributeAnnotation

__all__ = [
    "ParameterSpecification",
    "ParametersSpecification",
    "parameter_specification",
]


@final
class ParameterNoneSpecification(TypedDict, total=False):
    type: Required[Literal["null"]]
    description: NotRequired[str]


@final
class ParameterBoolSpecification(TypedDict, total=False):
    type: Required[Literal["boolean"]]
    description: NotRequired[str]


@final
class ParameterIntegerSpecification(TypedDict, total=False):
    type: Required[Literal["integer"]]
    description: NotRequired[str]


@final
class ParameterNumberSpecification(TypedDict, total=False):
    type: Required[Literal["number"]]
    description: NotRequired[str]


@final
class ParameterStringSpecification(TypedDict, total=False):
    type: Required[Literal["string"]]
    description: NotRequired[str]
    format: NotRequired[
        Literal[
            "uri",
            "uuid",
            "date",
            "time",
            "date-time",
        ]
    ]


@final
class ParameterStringEnumSpecification(TypedDict, total=False):
    type: NotRequired[Literal["string"]]
    enum: Required[Sequence[str]]
    description: NotRequired[str]


@final
class ParameterIntegerEnumSpecification(TypedDict, total=False):
    type: NotRequired[Literal["integer"]]
    enum: Required[Sequence[int]]
    description: NotRequired[str]


@final
class ParameterNumberEnumSpecification(TypedDict, total=False):
    type: NotRequired[Literal["number"]]
    enum: Required[Sequence[float]]
    description: NotRequired[str]


ParameterEnumSpecification = (
    ParameterStringEnumSpecification
    | ParameterIntegerEnumSpecification
    | ParameterNumberEnumSpecification
)


@final
class ParameterUnionSpecification(TypedDict, total=False):
    oneOf: Required[Sequence["ParameterSpecification"]]
    description: NotRequired[str]


@final
class ParameterArraySpecification(TypedDict, total=False):
    type: Required[Literal["array"]]
    items: NotRequired["ParameterSpecification"]
    description: NotRequired[str]


@final
class ParameterTupleSpecification(TypedDict, total=False):
    type: Required[Literal["array"]]
    prefixItems: Required[Sequence["ParameterSpecification"]]
    description: NotRequired[str]


@final
class ParameterDictSpecification(TypedDict, total=False):
    type: Required[Literal["object"]]
    additionalProperties: Required["ParameterSpecification | bool"]
    description: NotRequired[str]
    required: NotRequired[Sequence[str]]


@final
class ParameterObjectSpecification(TypedDict, total=False):
    type: Required[Literal["object"]]
    properties: Required[Mapping[str, "ParameterSpecification"]]
    title: NotRequired[str]
    description: NotRequired[str]
    required: NotRequired[Sequence[str]]


@final
class ParameterAnyObjectSpecification(TypedDict, total=False):
    type: Required[Literal["object"]]
    additionalProperties: Required[Literal[True]]
    description: NotRequired[str]


ReferenceParameterSpecification = TypedDict(
    "ReferenceParameterSpecification",
    {
        "$ref": Required[str],
        "description": NotRequired[str],
    },
    total=False,
)


type ParameterSpecification = (
    ParameterNoneSpecification
    | ParameterBoolSpecification
    | ParameterIntegerSpecification
    | ParameterNumberSpecification
    | ParameterStringSpecification
    | ParameterEnumSpecification
    | ParameterUnionSpecification
    | ParameterArraySpecification
    | ParameterTupleSpecification
    | ParameterDictSpecification
    | ParameterObjectSpecification
    | ParameterAnyObjectSpecification
    | ReferenceParameterSpecification
)

type ParametersSpecification = ParameterObjectSpecification | ParameterAnyObjectSpecification


def parameter_specification(  # noqa: PLR0911
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if specification := SPECIFICATIONS.get(annotation.origin):
        return specification(annotation, description)

    elif hasattr(annotation.origin, "__PARAMETERS_SPECIFICATION__"):
        if description := description:
            return cast(
                ParameterSpecification,
                {
                    **annotation.origin.__PARAMETERS_SPECIFICATION__,
                    "description": description,
                },
            )

        else:
            return annotation.origin.__PARAMETERS_SPECIFICATION__

    elif is_typeddict(annotation.origin):
        return _prepare_specification_of_typed_dict(
            annotation,
            description=description,
        )

    elif issubclass(annotation.origin, IntEnum):
        if description := description:
            return {
                "type": "integer",
                "enum": list(annotation.origin),
                "description": description,
            }

        else:
            return {
                "type": "integer",
                "enum": list(annotation.origin),
            }

    elif issubclass(annotation.origin, StrEnum):
        if description := description:
            return {
                "type": "string",
                "enum": list(annotation.origin),
                "description": description,
            }

        else:
            return {
                "type": "string",
                "enum": list(annotation.origin),
            }

    else:
        raise TypeError(f"Unsupported type annotation: {annotation}")


def _prepare_specification_of_any(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "object",
            "additionalProperties": True,
            "description": description,
        }

    else:
        return {
            "type": "object",
            "additionalProperties": True,
        }


def _prepare_specification_of_none(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "null",
            "description": description,
        }

    else:
        return {
            "type": "null",
        }


def _prepare_specification_of_missing(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "null",
            "description": description,
        }

    else:
        return {
            "type": "null",
        }


def _prepare_specification_of_literal(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if all(isinstance(element, str) for element in annotation.arguments):
        if description := description:
            return {
                "type": "string",
                "enum": annotation.arguments,
                "description": description,
            }

        else:
            return {
                "type": "string",
                "enum": annotation.arguments,
            }
    else:
        raise TypeError(f"Unsupported type annotation: {annotation}")


def _prepare_specification_of_sequence(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "array",
            "items": parameter_specification(
                annotation.arguments[0],
                description=None,
            ),
            "description": description,
        }

    else:
        return {
            "type": "array",
            "items": parameter_specification(
                annotation.arguments[0],
                description=None,
            ),
        }


def _prepare_specification_of_mapping(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "object",
            "additionalProperties": parameter_specification(
                annotation.arguments[1],
                description=None,
            ),
            "description": description,
        }

    else:
        return {
            "type": "object",
            "additionalProperties": parameter_specification(
                annotation.arguments[1],
                description=None,
            ),
        }


def _prepare_specification_of_tuple(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if (
        annotation.arguments[-1].origin == Ellipsis
        or annotation.arguments[-1].origin == EllipsisType
    ):
        if description := description:
            return {
                "type": "array",
                "items": parameter_specification(annotation.arguments[0], description=None),
                "description": description,
            }

        else:
            return {
                "type": "array",
                "items": parameter_specification(annotation.arguments[0], description=None),
            }

    elif description := description:
        return {
            "type": "array",
            "prefixItems": [
                parameter_specification(element, description=None)
                for element in annotation.arguments
            ],
            "description": description,
        }

    else:
        return {
            "type": "array",
            "prefixItems": [
                parameter_specification(element, description=None)
                for element in annotation.arguments
            ],
        }


def _prepare_specification_of_union(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "oneOf": [
                parameter_specification(cast(AttributeAnnotation, argument), description=None)
                for argument in annotation.arguments
            ],
            "description": description,
        }

    else:
        return {
            "oneOf": [
                parameter_specification(cast(AttributeAnnotation, argument), description=None)
                for argument in annotation.arguments
            ],
        }


def _prepare_specification_of_bool(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "boolean",
            "description": description,
        }

    else:
        return {
            "type": "boolean",
        }


def _prepare_specification_of_int(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "integer",
            "description": description,
        }

    else:
        return {
            "type": "integer",
        }


def _prepare_specification_of_float(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "number",
            "description": description,
        }

    else:
        return {
            "type": "number",
        }


def _prepare_specification_of_str(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "string",
            "description": description,
        }

    else:
        return {
            "type": "string",
        }


def _prepare_specification_of_uuid(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "string",
            "format": "uuid",
            "description": description,
        }

    else:
        return {
            "type": "string",
            "format": "uuid",
        }


def _prepare_specification_of_date(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "string",
            "format": "date",
            "description": description,
        }

    else:
        return {
            "type": "string",
            "format": "date",
        }


def _prepare_specification_of_datetime(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "string",
            "format": "date-time",
            "description": description,
        }

    else:
        return {
            "type": "string",
            "format": "date-time",
        }


def _prepare_specification_of_time(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "string",
            "format": "time",
            "description": description,
        }

    else:
        return {
            "type": "string",
            "format": "time",
        }


def _prepare_specification_of_typed_dict(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    required: list[str] = []
    properties: dict[str, ParameterSpecification] = {}

    for key, element in annotation.extra.items():
        properties[key] = parameter_specification(
            element,
            description=None,
        )

        if getattr(element, "required", True):
            required.append(key)

    if description := description:
        return {
            "type": "object",
            "properties": properties,
            "required": required,
            "description": description,
        }

    else:
        return {
            "type": "object",
            "properties": properties,
            "required": required,
        }


SPECIFICATIONS: Mapping[
    Any, Callable[[AttributeAnnotation, str | None], ParameterSpecification]
] = {
    Any: _prepare_specification_of_any,
    NoneType: _prepare_specification_of_none,
    Missing: _prepare_specification_of_missing,
    bool: _prepare_specification_of_bool,
    int: _prepare_specification_of_int,
    float: _prepare_specification_of_float,
    str: _prepare_specification_of_str,
    tuple: _prepare_specification_of_tuple,
    Literal: _prepare_specification_of_literal,
    Sequence: _prepare_specification_of_sequence,
    SequenceType: _prepare_specification_of_sequence,
    Mapping: _prepare_specification_of_mapping,
    MappingType: _prepare_specification_of_mapping,
    UUID: _prepare_specification_of_uuid,
    date: _prepare_specification_of_date,
    datetime: _prepare_specification_of_datetime,
    time: _prepare_specification_of_time,
    Union: _prepare_specification_of_union,
    UnionType: _prepare_specification_of_union,
}



from collections.abc import Sequence

from haiway import MISSING, Missing, State

__all__ = [
    "BedrockChatConfig",
]


class BedrockChatConfig(State):
    model: str
    temperature: float = 1.0
    top_p: float | Missing = MISSING
    max_tokens: int | Missing = MISSING
    timeout: float | Missing = MISSING
    stop_sequences: Sequence[str] | Missing = MISSING



from collections.abc import Callable, Iterable, Mapping
from typing import Any, Protocol, Self, overload, runtime_checkable

from haiway import State

from draive.instructions import Instruction
from draive.lmm import LMMCompletion, LMMInput, LMMOutputSelection
from draive.multimodal import Multimodal, MultimodalContent
from draive.prompts import Prompt
from draive.tools import AnyTool, Toolbox

__all__ = [
    "Step",
    "StepsCompleting",
]


@runtime_checkable
class StepResultProcessing(Protocol):
    async def __call__(
        self,
        content: MultimodalContent,
    ) -> MultimodalContent: ...


class Step(State):
    @overload
    @classmethod
    def of(
        cls,
        input: Prompt | Multimodal,
        /,
        *,
        completion: Multimodal,
        volatile: bool = False,
        extends_result: bool = False,
        condition: Callable[[], bool] | None = None,
        **extra: Any,
    ) -> Self: ...

    @overload
    @classmethod
    def of(
        cls,
        input: Prompt | Multimodal,
        /,
        *,
        instruction: Instruction | str | None = None,
        tools: Toolbox | Iterable[AnyTool] | None = None,
        output: LMMOutputSelection = "auto",
        volatile: bool = False,
        extends_result: bool = False,
        condition: Callable[[], bool] | None = None,
        result_processing: StepResultProcessing | None = None,
        **extra: Any,
    ) -> Self: ...

    @classmethod
    def of(  # noqa: PLR0913
        cls,
        input: Prompt | Multimodal,  # noqa: A002
        /,
        *,
        instruction: Instruction | str | None = None,
        tools: Toolbox | Iterable[AnyTool] | None = None,
        output: LMMOutputSelection = "auto",
        completion: Multimodal | None = None,
        volatile: bool = False,
        extends_result: bool = False,
        condition: Callable[[], bool] | None = None,
        result_processing: StepResultProcessing | None = None,
        **extra: Any,
    ) -> Self:
        assert (  # nosec: B101
            completion is None or instruction is None
        ), "Can't specify instruction with predefined result"
        assert (  # nosec: B101
            completion is None or tools is None
        ), "Can't specify tools with predefined result"
        assert (  # nosec: B101
            completion is None or result_processing is None
        ), "Can't specify result processing with predefined result"

        step_input: Prompt | LMMInput
        match input:
            case Prompt() as prompt:
                step_input = prompt

            case other:
                step_input = LMMInput.of(other)

        return cls(
            instruction=Instruction.of(instruction) if instruction else None,
            input=step_input,
            toolbox=Toolbox.out_of(tools),
            output=output,
            completion=None if completion is None else LMMCompletion.of(completion),
            volatile=volatile,
            extends_result=extends_result,
            condition=condition,
            result_processing=result_processing,
            extra=extra,
        )

    instruction: Instruction | None
    input: Prompt | LMMInput
    toolbox: Toolbox
    output: LMMOutputSelection
    completion: LMMCompletion | None
    volatile: bool
    extends_result: bool
    condition: Callable[[], bool] | None
    result_processing: StepResultProcessing | None
    extra: Mapping[str, Any]


@runtime_checkable
class StepsCompleting(Protocol):
    async def __call__(
        self,
        *steps: Step | Prompt | Multimodal,
        instruction: Instruction | str | None = None,
        **extra: Any,
    ) -> MultimodalContent: ...



from draive.generation.text.call import generate_text
from draive.generation.text.state import TextGeneration
from draive.generation.text.types import TextGenerator

__all__ = [
    "TextGeneration",
    "TextGenerator",
    "generate_text",
]



from draive.lmm.call import lmm_invoke, lmm_stream
from draive.lmm.state import LMMInvocation, LMMStream
from draive.lmm.types import (
    LMMCompletion,
    LMMContext,
    LMMContextElement,
    LMMInput,
    LMMInvocating,
    LMMOutput,
    LMMOutputSelection,
    LMMStreamChunk,
    LMMStreaming,
    LMMStreamInput,
    LMMStreamOutput,
    LMMStreamProperties,
    LMMToolException,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
    LMMToolResponses,
    LMMToolSelection,
    LMMToolSpecification,
)

__all__ = [
    "LMMCompletion",
    "LMMContext",
    "LMMContextElement",
    "LMMInput",
    "LMMInvocating",
    "LMMInvocation",
    "LMMOutput",
    "LMMOutputSelection",
    "LMMStream",
    "LMMStreamChunk",
    "LMMStreamInput",
    "LMMStreamOutput",
    "LMMStreamProperties",
    "LMMStreaming",
    "LMMToolException",
    "LMMToolRequest",
    "LMMToolRequests",
    "LMMToolResponse",
    "LMMToolResponses",
    "LMMToolSelection",
    "LMMToolSpecification",
    "lmm_invoke",
    "lmm_stream",
]



from draive.ollama.client import OllamaClient
from draive.ollama.config import OllamaChatConfig
from draive.ollama.lmm import ollama_lmm
from draive.ollama.types import OllamaException

__all__ = [
    "OllamaChatConfig",
    "OllamaClient",
    "OllamaException",
    "ollama_lmm",
]



from typing import Any

from haiway import ctx

from draive.tokenization.state import Tokenization

__all__ = [
    "count_text_tokens",
    "tokenize_text",
]


def tokenize_text(
    text: str,
    **extra: Any,
) -> list[int]:
    """
    Tokenize input text using current Tokenization state.
    """

    return ctx.state(Tokenization).tokenize_text(
        text=text,
        **extra,
    )


def count_text_tokens(
    text: str,
    **extra: Any,
) -> int:
    """
    Count input text tokens using current Tokenization state.
    """

    return len(
        ctx.state(Tokenization).tokenize_text(
            text=text,
            **extra,
        )
    )



from collections.abc import Sequence
from typing import cast

from draive.embedding import Embedded, embed_images, embed_texts
from draive.evaluation import EvaluationScore, EvaluationScoreValue, evaluator
from draive.multimodal import MediaContent, Multimodal, MultimodalContent, MultimodalTagElement
from draive.similarity.score import vector_similarity_score
from draive.steps import steps_completion

__all__ = [
    "image_vector_similarity_evaluator",
    "similarity_evaluator",
    "text_vector_similarity_evaluator",
]


INSTRUCTION: str = """\
You are evaluating the provided content according to the defined criteria.

<INSTRUCTION>
Compare the REFERENCE and the EVALUATED content by carefully examining them, then rate\
 the EVALUATED content using solely a similarity metric according to the EVALUATION_CRITERIA.
Think step by step and provide explanation of the score before the final score.
Use the explained RATING scale and the requested FORMAT to provide the result.
</INSTRUCTION>

<EVALUATION_CRITERIA>
Evaluated metric is similarity - the degree of semantic similarity between the REFERENCE\
 and the EVALUATED content.
</EVALUATION_CRITERIA>
{guidelines}
<RATING>
Assign a similarity score using exact name of one of the following values:
- "poor" is very low similarity, the content is completely unrelated in meaning.
- "good" is moderate similarity, the content share some common themes or ideas.
- "perfect" is very high similarity, the content is very close in meaning\
 or convey the same information.
Use the "none" value for content that cannot be rated at all.
</RATING>

<FORMAT>
The final result containing only the numerical score value HAVE to be put inside a `RESULT` \
xml tag within the result i.e. `<RESULT>good</RESULT>`.
</FORMAT>
"""


@evaluator(name="similarity")
async def similarity_evaluator(
    evaluated: Multimodal,
    /,
    *,
    reference: Multimodal,
    guidelines: str | None = None,
) -> EvaluationScore:
    if not evaluated:
        return EvaluationScore(
            value=0,
            comment="Input was empty!",
        )

    if not reference:
        return EvaluationScore(
            value=0,
            comment="Reference was empty!",
        )

    completion: MultimodalContent = await steps_completion(
        MultimodalContent.of(
            "<REFERENCE>",
            reference,
            "</REFERENCE>\n<EVALUATED>",
            evaluated,
            "</EVALUATED>",
        ),
        instruction=INSTRUCTION.format(
            guidelines=f"\n<GUIDELINES>\n{guidelines}\n</GUIDELINES>\n"
            if guidelines is not None
            else ""
        ),
    )

    if result := MultimodalTagElement.parse_first(
        completion,
        tag="RESULT",
    ):
        return EvaluationScore.of(
            cast(EvaluationScoreValue, result.content.as_string()),
            comment=completion.as_string(),
        )

    else:
        raise ValueError("Invalid evaluator result")


@evaluator(name="text_vector_similarity")
async def text_vector_similarity_evaluator(
    evaluated: str,
    /,
    *,
    reference: str,
) -> float:
    embedding: Sequence[Embedded[str]] = await embed_texts([reference, evaluated])

    return vector_similarity_score(
        value_vector=embedding[0].vector,
        reference_vector=embedding[1].vector,
    )


@evaluator(name="image_vector_similarity")
async def image_vector_similarity_evaluator(
    evaluated: MediaContent | bytes,
    /,
    *,
    reference: MediaContent | bytes,
) -> float:
    evaluated_data: bytes
    match evaluated:
        case MediaContent() as media:
            match media.source:
                case bytes() as data:
                    evaluated_data = data

                case str():
                    raise ValueError("Unsupported media source")

        case raw_data:
            evaluated_data = raw_data

    reference_data: bytes
    match reference:
        case MediaContent() as media:
            match media.source:
                case bytes() as data:
                    reference_data = data

                case str():
                    raise ValueError("Unsupported media source")

        case raw_data:
            reference_data = raw_data

    embedding: Sequence[Embedded[bytes]] = await embed_images([reference_data, evaluated_data])

    return vector_similarity_score(
        value_vector=embedding[0].vector,
        reference_vector=embedding[1].vector,
    )



from collections.abc import Callable, Coroutine
from typing import final, overload

from haiway import ArgumentsTrace, ResultTrace, ctx, freeze

from draive.instructions.types import (
    Instruction,
    InstructionDeclaration,
    InstructionDeclarationArgument,
)
from draive.parameters import ParametrizedFunction

__all__ = [
    "InstructionTemplate",
    "instruction",
]


@final
class InstructionTemplate[**Args](ParametrizedFunction[Args, Coroutine[None, None, str]]):
    def __init__(
        self,
        /,
        name: str,
        *,
        description: str | None = None,
        function: Callable[Args, Coroutine[None, None, str]],
    ) -> None:
        super().__init__(function)

        self.declaration: InstructionDeclaration = InstructionDeclaration(
            name=name,
            description=description,
            arguments=[
                InstructionDeclarationArgument(
                    name=parameter.alias or parameter.name,
                    specification=parameter.specification,
                    required=parameter.required,
                )
                for parameter in self._parameters.values()
            ],
        )

        freeze(self)

    async def resolve(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Instruction:
        with ctx.scope(f"instruction:{self.declaration.name}"):
            ctx.record(ArgumentsTrace.of(*args, **kwargs))
            try:
                result = Instruction(
                    name=self.declaration.name,
                    description=self.declaration.description,
                    content=await super().__call__(*args, **kwargs),  # pyright: ignore[reportCallIssue]
                )
                ctx.record(ResultTrace.of(result))

                return result

            except BaseException as exc:
                ctx.record(ResultTrace.of(exc))
                ctx.log_error(
                    "Instruction resolving error",
                    exception=exc,
                )
                raise exc


@overload
def instruction[**Args](
    function: Callable[Args, Coroutine[None, None, str]],
    /,
) -> Callable[[Callable[Args, Coroutine[None, None, str]]], InstructionTemplate[Args]]: ...


@overload
def instruction[**Args](
    *,
    name: str | None = None,
    description: str | None = None,
) -> InstructionTemplate[Args]: ...


def instruction[**Args](
    function: Callable[Args, Coroutine[None, None, str]] | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
) -> (
    Callable[[Callable[Args, Coroutine[None, None, str]]], InstructionTemplate[Args]]
    | InstructionTemplate[Args]
):
    def wrap(
        function: Callable[Args, Coroutine[None, None, str]],
    ) -> InstructionTemplate[Args]:
        return InstructionTemplate[Args](
            name=name or function.__name__,
            description=description,
            function=function,
        )

    if function := function:
        return wrap(function=function)

    else:
        return wrap



from pathlib import Path

from draive.sentencepiece import sentencepiece_processor, sentencepiece_tokenizer
from draive.tokenization import Tokenization

__all__ = [
    "gemini_tokenizer",
]


async def gemini_tokenizer(
    model_name: str,
    /,
) -> Tokenization:
    return Tokenization(
        tokenize_text=sentencepiece_tokenizer(
            await sentencepiece_processor(model_path=_model_path(model_name))
        )
    )


def _model_path(model_name: str) -> str:
    model_file: str = _mapping.get(model_name, "gemini_tokenizer.model")
    return str(Path(__file__).parent / "tokens" / model_file)


_mapping: dict[str, str] = {
    "gemini-1.5-flash": "gemini_tokenizer.model",
    "gemini-1.5-pro": "gemini_tokenizer.model",
}



from draive.anthropic.client import AnthropicClient
from draive.anthropic.config import AnthropicConfig
from draive.anthropic.lmm import anthropic_lmm
from draive.anthropic.tokenization import anthropic_tokenizer
from draive.anthropic.types import AnthropicException

__all__ = [
    "AnthropicClient",
    "AnthropicConfig",
    "AnthropicException",
    "anthropic_lmm",
    "anthropic_tokenizer",
]



from draive.parameters.function import Argument, ParametrizedFunction
from draive.parameters.model import DataModel, Field
from draive.parameters.specification import ParameterSpecification, ParametersSpecification
from draive.parameters.types import BasicValue, ParameterValidationContext, ParameterValidationError
from draive.parameters.validation import ParameterValidator, ParameterVerification

__all__ = [
    "Argument",
    "BasicValue",
    "DataModel",
    "DataModel",
    "Field",
    "ParameterSpecification",
    "ParameterValidationContext",
    "ParameterValidationError",
    "ParameterValidator",
    "ParameterVerification",
    "ParametersSpecification",
    "ParametrizedFunction",
]



from collections.abc import Callable, Coroutine, Mapping
from typing import Any, Protocol, cast, final, overload

from haiway import ArgumentsTrace, ResultTrace, ctx, freeze

from draive.lmm.types import (
    LMMToolError,
    LMMToolException,
    LMMToolFunctionSpecification,
    LMMToolSpecification,
)
from draive.multimodal import Multimodal, MultimodalContent, MultimodalContentConvertible
from draive.parameters import ParameterSpecification, ParametersSpecification, ParametrizedFunction

__all__ = [
    "AnyTool",
    "Tool",
    "ToolAvailabilityCheck",
    "tool",
]


class ToolAvailabilityCheck(Protocol):
    def __call__(self) -> bool: ...


@final
class Tool[**Args, Result](ParametrizedFunction[Args, Coroutine[None, None, Result]]):
    def __init__(  # noqa: PLR0913
        self,
        /,
        name: str,
        *,
        function: Callable[Args, Coroutine[None, None, Result]],
        description: str | None = None,
        specification: ParametersSpecification | None = None,
        availability_check: ToolAvailabilityCheck | None = None,
        format_result: Callable[[Result], Multimodal],
        format_failure: Callable[[Exception], Multimodal],
        direct_result: bool = False,
    ) -> None:
        super().__init__(function)

        if specification is None:
            aliased_required: list[str] = []
            parameters: dict[str, ParameterSpecification] = {}
            for parameter in self._parameters.values():
                parameters[parameter.alias or parameter.name] = parameter.specification

                if parameter.required:
                    aliased_required.append(parameter.alias or parameter.name)

            specification = {
                "type": "object",
                "properties": parameters,
                "required": aliased_required,
            }

        self.specification: LMMToolSpecification = LMMToolFunctionSpecification(
            name=name,
            description=description,
            parameters=specification,
        )

        self.name: str = name
        self.description: str | None = description
        self._direct_result: bool = direct_result
        self._check_availability: ToolAvailabilityCheck = availability_check or (
            lambda: True  # available by default
        )
        self.format_result: Callable[[Result], Multimodal] = format_result
        self.format_failure: Callable[[Exception], Multimodal] = format_failure

        freeze(self)

    @property
    def available(self) -> bool:
        try:
            return self._check_availability()

        except Exception:
            return False

    @property
    def requires_direct_result(self) -> bool:
        return self._direct_result

    # call from toolbox
    async def _toolbox_call(
        self,
        call_id: str,
        /,
        arguments: Mapping[str, Any],
    ) -> MultimodalContent:
        with ctx.scope(self.name):
            ctx.record(ArgumentsTrace.of(**arguments))
            try:
                try:
                    if not self.available:
                        raise LMMToolException(f"{self.name} is not available!")

                    result: Result = await super().__call__(**arguments)  # pyright: ignore[reportCallIssue]
                    ctx.record(ResultTrace.of(result))

                    return MultimodalContent.of(self.format_result(result))

                except Exception as exc:
                    # return an error with formatted content
                    raise LMMToolError(
                        f"Tool {self.name}[{call_id}] failed",
                        content=MultimodalContent.of(self.format_failure(exc)),
                    ) from exc

            except BaseException as exc:
                ctx.record(ResultTrace.of(exc))
                ctx.log_error(
                    "Tool call error",
                    exception=exc,
                )
                raise exc

    # regular call when using as a function
    async def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        with ctx.scope(self.name):
            ctx.record(ArgumentsTrace.of(*args, **kwargs))
            try:
                result: Result = await super().__call__(
                    *args,
                    **kwargs,
                )
                ctx.record(ResultTrace.of(result))

                return result

            except BaseException as exc:
                ctx.record(ResultTrace.of(exc))
                ctx.log_error(
                    "Tool call error",
                    exception=exc,
                )
                raise exc


AnyTool = Tool[Any, Any]


class ToolWrapper(Protocol):
    def __call__[**Args, Result](
        self,
        function: Callable[Args, Coroutine[None, None, Result]],
    ) -> Tool[Args, Result]: ...


class PartialToolWrapper[Result](Protocol):
    def __call__[**Args](
        self,
        function: Callable[Args, Coroutine[None, None, Result]],
    ) -> Tool[Args, Result]: ...


@overload
def tool[**Args, Result](
    function: Callable[Args, Coroutine[None, None, Result]],
    /,
) -> Tool[Args, Result]:
    """
    Convert a function to a tool using default parameters and no description.

    In order to adjust the arguments behavior and specification use an instance of Argument
    as a default value of any given argument with desired configuration
    for each argument individually.

    Parameters
    ----------
    function: Callable[Args, Coroutine[None, None, Result]]
        a function to be wrapped as a Tool.
    Returns
    -------
    Tool[Args, Result]
        a Tool representation of the provided function.
    """


@overload
def tool[Result](
    *,
    name: str | None = None,
    description: str | None = None,
    availability_check: ToolAvailabilityCheck | None = None,
    format_result: Callable[[Result], Multimodal],
    format_failure: Callable[[Exception], Multimodal] | None = None,
    direct_result: bool = False,
) -> PartialToolWrapper[Result]:
    """
    Convert a function to a tool using provided parameters.

    In order to adjust the arguments behavior and specification use an instance of Argument
    as a default value of any given argument with desired configuration
    for each argument individually.

    Parameters
    ----------
    name: str
        name to be used in a tool specification.
        Default is the name of the wrapped function.
    description: int
        description to be used in a tool specification. Allows to present the tool behavior to the
        external system.
        Default is empty.
    availability_check: ToolAvailabilityCheck
        function used to verify availability of the tool in given context. It can be used to check
        permissions or occurrence of a specific state to allow its usage.
        Provided function should raise an Exception when the tool should not be available.
        Default is always available.
    format_result: Callable[[Result], MultimodalContent]
        function converting tool result to MultimodalContent. It is used to format the result
        for model processing. Default implementation converts the result to string if needed.
    format_failure: Callable[[Exception], MultimodalContent]
        function converting tool call exception to a fallback MultimodalContent.
        Default implementation return "ERROR" string and logs the exception.
    direct_result: bool
        controls if tool result should break the ongoing processing and be the direct result of it.
        Note that during concurrent execution of multiple tools the call/result order defines
        direct result and exact behavior is not defined.
        Default is False.

    Returns
    -------
    Callable[[Function[Args, Coroutine[None, None, Result]]], Tool[Args, Result]]
        function allowing to convert other function to a Tool using provided configuration.
    """


@overload
def tool(
    *,
    name: str | None = None,
    description: str | None = None,
    availability_check: ToolAvailabilityCheck | None = None,
    format_failure: Callable[[Exception], Multimodal] | None = None,
    direct_result: bool = False,
) -> ToolWrapper:
    """
    Convert a function to a tool using provided parameters.

    In order to adjust the arguments behavior and specification use an instance of Argument
    as a default value of any given argument with desired configuration
    for each argument individually.

    Parameters
    ----------
    name: str
        name to be used in a tool specification.
        Default is the name of the wrapped function.
    description: int
        description to be used in a tool specification. Allows to present the tool behavior to the
        external system.
        Default is empty.
    availability_check: ToolAvailabilityCheck
        function used to verify availability of the tool in given context. It can be used to check
        permissions or occurrence of a specific state to allow its usage.
        Provided function should raise an Exception when the tool should not be available.
        Default is always available.
    format_result: Callable[[Result], MultimodalContent]
        function converting tool result to MultimodalContent. It is used to format the result
        for model processing. Default implementation converts the result to string if needed.
    format_failure: Callable[[Exception], MultimodalContent]
        function converting tool call exception to a fallback MultimodalContent.
        Default implementation return "ERROR" string and logs the exception.
    direct_result: bool
        controls if tool result should break the ongoing processing and be the direct result of it.
        Note that during concurrent execution of multiple tools the call/result order defines
        direct result and exact behavior is not defined.
        Default is False.

    Returns
    -------
    Callable[[Function[Args, Coroutine[None, None, Result]]], Tool[Args, Result]]
        function allowing to convert other function to a Tool using provided configuration.
    """


def tool[**Args, Result](  # noqa: PLR0913
    function: Callable[Args, Coroutine[None, None, Result]] | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
    availability_check: ToolAvailabilityCheck | None = None,
    format_result: Callable[[Result], Multimodal] | None = None,
    format_failure: Callable[[Exception], Multimodal] | None = None,
    direct_result: bool = False,
) -> PartialToolWrapper[Result] | ToolWrapper | Tool[Args, Result]:
    def wrap(
        function: Callable[Args, Coroutine[None, None, Result]],
    ) -> Tool[Args, Result]:
        return Tool(
            name=name or function.__name__,
            description=description,
            function=function,
            availability_check=availability_check,
            format_result=format_result or _default_result_format,
            format_failure=format_failure or _default_failure_result,
            direct_result=direct_result,
        )

    if function := function:
        return wrap(function=function)

    else:
        return cast(PartialToolWrapper[Result], wrap)


def _default_result_format(result: Any) -> MultimodalContent:
    match result:
        case MultimodalContent() as content:
            return content

        case element if isinstance(element, MultimodalContentConvertible):
            return MultimodalContent.of(element)

        case [*elements]:
            return MultimodalContent.of(
                *[
                    element if isinstance(element, MultimodalContentConvertible) else str(element)
                    for element in elements
                ]
            )

        case other:
            return MultimodalContent.of(str(other))


def _default_failure_result(exception: Exception) -> MultimodalContent:
    return MultimodalContent.of("ERROR")
